# @package      hubzero-submit-client
# @file         SubmitClient.py
# @author       Steven Clark <clarks@purdue.edu>
# @copyright    Copyright (c) 2012-2015 HUBzero Foundation, LLC.
# @license      http://opensource.org/licenses/MIT MIT
#
# Copyright (c) 2012-2015 HUBzero Foundation, LLC.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# HUBzero is a registered trademark of HUBzero Foundation, LLC.
#
__version__ = '3.0.1'

import os
import sys
import re
import shutil
import shlex
import select
import socket
import time
import signal
import resource
import subprocess
import math
import random
import stat
import csv
import json
import logging
from errno import EPIPE

from hubzero.submit.LogMessage             import getLogIDMessage as getLogMessage, isLogMessageFileStdout, isLogMessageFileStderr, isLoggingOn
from hubzero.submit.CommandParser          import CommandParser
from hubzero.submit.ParameterTemplate      import ParameterTemplate
from hubzero.submit.LocalBatchAppScript    import LocalBatchAppScript
from hubzero.submit.JobScanner             import JobScanner
from hubzero.submit.JobStatistic           import JobStatistic
from hubzero.submit.TarCommand             import buildCreate as buildCreateTarCommand
from hubzero.submit.UnboundConnection      import UnboundConnection
from hubzero.submit.ClientIdAuthAttributes import ClientIdAuthAttributes
from hubzero.submit.SubmissionScriptsInfo  import SubmissionScriptsInfo

TIMESTAMPTRANSFERRED = ".__timestamp_transferred"
TIMESTAMPFINISH      = ".__timestamp_finish"
TIMESTAMPSTART       = ".__timestamp_start"
TIMERESULTS          = ".__time_results"
EXITCODE             = ".__exit_code"

class SubmitClient:
   AUTHORIZEDACTIONNONE        = 0
   AUTHORIZEDACTIONEXECUTE     = 1 << 0
   AUTHORIZEDACTIONHEARTBEAT   = 1 << 1
   AUTHORIZEDACTIONSENDMETRICS = 1 << 2

   def __init__(self,
                configFilePath):
      self.logger         = logging.getLogger(__name__)
      self.configData     = {}
      self.configFilePath = configFilePath

      self.clientIdAuthAttributes = ClientIdAuthAttributes()
      self.serverConnection       = None
      self.commandParser          = None
      self.localWorkflowPEGASUS   = None
      self.jobScanner             = None

      self.serverData                      = {}
      self.serverData['version']           = ""
      self.serverData['authz']             = False
      self.serverData['jobId']             = 0
      self.serverData['jobToken']          = ""
      self.serverData['submitterClass']    = 1
      self.serverData['event']             = ""
      self.serverData['createdSessions']   = []
      self.serverData['heartbeatInterval'] = 0
      self.serverData['pegasusRC']         = ""
      self.serverData['pegasusSites']      = ""
      self.serverData['submissionScripts'] = None
      self.serverData['childHasExited']    = False
      self.serverData['exited']            = False

      self.clientData                        = {}
      self.clientData['hostFQDN']            = socket.getfqdn()
      self.clientData['version']             = __version__
      self.clientData['operationMode']       = 0
      self.clientData['authorizedAction']    = self.AUTHORIZEDACTIONNONE
      self.clientData['localJobId']          = "00000000"
      self.clientData['userName']            = None
      self.clientData['userId']              = None
      self.clientData['detach']              = False
      self.clientData['attachId']            = 0L
      self.clientData['pegasusExists']       = False
      self.clientData['pegasusVersion']      = None
      self.clientData['pegasusHome']         = None
      self.clientData['pegasusSetupError']   = ""
      self.clientData['reportMetrics']       = False
#     self.clientData['isClientTTY']         = True
      self.clientData['isClientTTY']         = sys.stdin.isatty() and sys.stdout.isatty()
      self.clientData['progressReport']      = 'text'
      self.clientData['readyToExit']         = False
      self.clientData['selectTimeout']       = 1*60.
      self.clientData['selectTimeoutSpread'] = 0.15
      self.clientData['disconnectMax']       = 3*self.clientData['selectTimeout']
      self.clientData['suspended']           = False
      self.clientData['transferFiles']       = []
      self.clientData['commandFiles']        = []
      self.clientData['inputObjects']        = {}
      self.clientData['outputObjects']       = {}
      self.clientData['isParametric']        = False
      self.clientData['localExecution']      = False
      self.clientData['isLocalPegasus']      = False
      self.clientData['runProgressMessage']  = ""
      self.clientData['childPid']            = None
      self.clientData['childReadStdout']     = None
      self.clientData['childReadStderr']     = None
      self.clientData['localStartTime']      = time.time()
      self.clientData['localFinishTime']     = None
      self.clientData['childHasExited']      = False
      self.clientData['childKillPending']    = False
      self.clientData['exitCode']            = 2
      self.clientData['cpuTime']             = 0.
      self.clientData['realTime']            = 0.
      self.clientData['jobStates']           = {}
      self.clientData['jobScanPath']         = ""
      self.clientData['jobScannerStarted']   = False
      self.clientData['runStatusPath']       = ""
      self.clientData['enteredCommand']      = ""
      self.clientData['startDate']           = None
      self.clientData['finishDate']          = None
      self.clientData['runName']             = ""
      self.clientData['jobPath']             = os.getcwd()
      self.clientData['importFilesComplete'] = False
      self.clientData['mountPoints']         = {}

      dfCommand = ['df','--portability','--all']
      child = subprocess.Popen(dfCommand,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               close_fds=True)
      dfStdOutput,dfStdError = child.communicate()
      dfExitStatus = child.returncode
      if dfExitStatus == 0:
         mounts = ''.join(dfStdOutput).strip().split('\n')[1:]
         for mount in mounts:
            device     = mount.split()[0]
            mountPoint = mount.split()[-1]
            if ':' in device:
               self.clientData['mountPoints'][mountPoint] = {'device':device,
                                                             'inputFiles':[],
                                                             'submitCommandFiles':[]}
            else:
               self.clientData['mountPoints'][mountPoint] = {'device':':'.join([self.clientData['hostFQDN'],device]),
                                                             'inputFiles':[],
                                                             'submitCommandFiles':[]}
      else:
         self.logger.log(logging.ERROR,getLogMessage("df command failed:\n%s\n" % (dfStdError)))
#     log(self.clientData['mountPoints'])
      self.clientData['workDirectory']       = os.getcwd()
      self.clientData['workDirectoryInode']  = os.lstat(self.clientData['workDirectory']).st_ino
      self.clientData['workDirectoryDevice'] = None
      mountPoint = self.clientData['workDirectory']
      while mountPoint:
         if os.path.ismount(mountPoint):
            try:
               self.clientData['workDirectoryDevice'] = self.clientData['mountPoints'][mountPoint]['device']
            except OSError,err:
               self.__writeToStderr("%s: %s\n" % (self.clientData['workDirectory'],err.args[1]))
               self.exit(err.args[0])
            except KeyError,err:
               self.__writeToStderr("Could not find mount point: %s\n" % (self.clientData['workDirectory']))
               self.exit(1)
            break
         else:
            if mountPoint != os.sep:
               mountPoint = os.path.dirname(mountPoint)
            else:
               break
      self.clientData['importObjects']       = {}
      self.clientData['exportObjects']       = {}
      self.clientData['closePendingObjects'] = []
      self.clientData['filesToRemove']       = []
      self.clientData['emptyFilesToRemove']  = []

      signal.signal(signal.SIGINT,self.handleTerminationSignal)
      signal.signal(signal.SIGHUP,self.handleTerminationSignal)
      signal.signal(signal.SIGQUIT,self.handleDetachSignal)
      signal.signal(signal.SIGABRT,self.handleTerminationSignal)
      signal.signal(signal.SIGTERM,self.handleTerminationSignal)
      signal.signal(signal.SIGCONT,self.handleContinueSignal)
      signal.signal(signal.SIGTSTP,self.handleSuspendSignal)
      signal.signal(signal.SIGXCPU,self.handleTerminationSignal)
      signal.signal(signal.SIGCHLD,self.handleLocalCompleteSignal)


   def __writeToStdout(self,
                       message):
      try:
         sys.stdout.write(message)
         sys.stdout.flush()
      except IOError,err:
         if not err[0] in [EPIPE]:
            self.logger.log(logging.ERROR,getLogMessage("Can't write to stdout: %s" % (message)))


   def __writeToStderr(self,
                       message):
      try:
         sys.stderr.write(message)
         sys.stderr.flush()
      except IOError,err:
         if not err[0] in [EPIPE]:
            self.logger.log(logging.ERROR,getLogMessage("Can't write to stderr: %s" % (message)))


   def sendUserErrorNotification(self,
                                 message):
      if   isLogMessageFileStdout():
         self.__writeToStdout(message + '\n')
      elif isLogMessageFileStderr():
         self.__writeToStderr(message + '\n')
      elif isLoggingOn():
         self.logger.log(logging.ERROR,getLogMessage(message))
      else:
         self.__writeToStderr(message + '\n')


   def killChild(self):
      if self.clientData['childPid'] and not self.clientData['childKillPending']:
         self.clientData['childKillPending'] = True
         try:
            os.kill(self.clientData['childPid'],signal.SIGINT)
            for retry in range(0,12):
               time.sleep(10)
               os.kill(self.clientData['childPid'],0)
               if not self.clientData['childKillPending']:
                  break
            if self.clientData['childKillPending']:
               os.kill(self.clientData['childPid'],signal.SIGTERM)
               for retry in range(0,6):
                  time.sleep(10)
                  os.kill(self.clientData['childPid'],0)
                  if not self.clientData['childKillPending']:
                     break
            if self.clientData['childKillPending']:
               os.kill(self.clientData['childPid'],signal.SIGKILL)
         except:
            pass


   def handleSuspendSignal(self,
                           signalNumber,
                           frame):
      clientMessage = {'messageType':'clientSuspended'}
      self.serverConnection.postJsonMessage(clientMessage)


   def handleContinueSignal(self,
                           signalNumber,
                           frame):
      clientMessage = {'messageType':'clientContinued'}
      self.serverConnection.postJsonMessage(clientMessage)


   def handleTerminationSignal(self,
                               signalNumber,
                               frame):
      if not self.clientData['readyToExit']:
         if self.clientData['localExecution']:
            if self.clientData['childPid']:
               self.killChild()
            else:
               exitCode = (1 << 7) | signalNumber
               self.clientData['exitCode'] = exitCode
         else:
            if self.serverConnection and self.serverConnection.isConnected():
               clientMessage = {'messageType':'clientSignal',
                                'signal':signalNumber}
               self.serverConnection.postJsonMessage(clientMessage)
            else:
               self.clientData['readyToExit'] = True
               exitCode = (1 << 7) | signalNumber
               self.clientData['exitCode'] = exitCode
               self.exit(exitCode)


   def handleLocalCompleteSignal(self,
                                 signalNumber,
                                 frame):
      self.clientData['childKillPending'] = False


   def handleDetachSignal(self,
                          signalNumber,
                          frame):
      if self.clientData['operationMode'] & self.commandParser.OPERATIONMODERUNDISTRIBUTORID:
         if not self.clientData['localExecution']:
            clientMessage = {'messageType':'detachSignal'}
            self.serverConnection.postJsonMessage(clientMessage)


   def setupPegasus(self):
      pegasusVersion    = None
      pegasusHome       = None
      pegasusPythonPath = None

      try:
         from Pegasus.DAX3 import ADAG

         pegasusCommand = ['pegasus-version']
         child = subprocess.Popen(pegasusCommand,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  close_fds=True)
         pegasusStdOutput,pegasusStdError = child.communicate()
         pegasusExitStatus = child.returncode
         if pegasusExitStatus == 0:
            pegasusVersion = pegasusStdOutput.strip()

            pegasusCommand = ['pegasus-config','--bin']
            child = subprocess.Popen(pegasusCommand,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE,
                                     close_fds=True)
            pegasusStdOutput,pegasusStdError = child.communicate()
            pegasusExitStatus = child.returncode
            if pegasusExitStatus == 0:
               pegasusHome = os.path.dirname(pegasusStdOutput.strip())

               pegasusCommand = ['pegasus-config','--python']
               child = subprocess.Popen(pegasusCommand,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                        close_fds=True)
               pegasusStdOutput,pegasusStdError = child.communicate()
               pegasusExitStatus = child.returncode
               if pegasusExitStatus == 0:
                  pegasusPythonPath = pegasusStdOutput.strip()
      except:
         pass

      if not pegasusVersion or not pegasusHome or not pegasusPythonPath:
         try:
            pegasusCommand = ". %s\n" % (self.configData['useSetup']) + \
                             "use -e -r pegasus-%s\n" % (self.configData['pegasusVersion']) + \
                             "pegasus-version\n" + \
                             "pegasus-config --bin\n" + \
                             "pegasus-config --python"
            child = subprocess.Popen(pegasusCommand,shell=True,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE,
                                     close_fds=True)
            pegasusStdOutput,pegasusStdError = child.communicate()
            pegasusExitStatus = child.returncode
            if pegasusExitStatus == 0:
               try:
                  pegasusVersion,pegasusBin,pegasusPythonPath = ''.join(pegasusStdOutput).strip().split()
                  pegasusHome = os.path.dirname(pegasusBin)
                  if not pegasusPythonPath in sys.path:
                     sys.path.insert(0,pegasusPythonPath)
                     from Pegasus.DAX3 import ADAG
               except:
                  pegasusVersion    = ""
                  pegasusHome       = ""
                  pegasusPythonPath = ""
            else:
               self.clientData['pegasusSetupError'] = "Pegasus version determination failed.\n%s\n" % (pegasusStdError)
         except:
            pegasusVersion    = ""
            pegasusHome       = ""
            pegasusPythonPath = ""
            self.clientData['pegasusSetupError'] = "Pegasus version determination failed.\n"

      if pegasusVersion and pegasusHome and pegasusPythonPath:
         self.clientData['pegasusExists'] = True
      self.clientData['pegasusVersion']   = pegasusVersion
      self.clientData['pegasusHome']      = pegasusHome


   def scheduleHeartbeat(self):
      signal.signal(signal.SIGALRM,self.heartbeat)
      signal.alarm(self.serverData['heartbeatInterval'])


   def heartbeat(self,
                 signalNumber,
                 frame):
      self.logger.log(logging.INFO,getLogMessage("Client heartbeat"))
      self.connect()
      if self.clientData['workDirectoryDevice']:
         clientMessage = {'messageType':'pwdWithMount',
                          'path':self.clientData['workDirectory'],
                          'inode':self.clientData['workDirectoryInode'],
                          'device':self.clientData['workDirectoryDevice']}
         self.serverConnection.postJsonMessage(clientMessage)
      else:
         clientMessage = {'messageType':'pwd',
                          'path':self.clientData['workDirectory'],
                          'inode':self.clientData['workDirectoryInode']}
         self.serverConnection.postJsonMessage(clientMessage)
      clientMessage = {'messageType':'jobId',
                       'jobId':self.serverData['jobId']}
      self.serverConnection.postJsonMessage(clientMessage)
      self.clientData['authorizedAction'] = self.AUTHORIZEDACTIONHEARTBEAT
      self.signon()


   def cancelHeartbeat(self):
      alarmFunction = signal.getsignal(signal.SIGALRM)
      if not alarmFunction in [None,signal.SIG_IGN,signal.SIG_DFL]:
         if cmp(alarmFunction,self.heartbeat) == 0:
            signal.alarm(0)


   def configure(self):
      sectionPattern  = re.compile('(\s*\[)([^\s]*)(]\s*)')
      keyValuePattern = re.compile('( *)(\w*)( *= *)(.*[^\s$])( *)')
      commentPattern  = re.compile('\s*#.*')
      inClientSection = False

      try:
         fpConfig = open(self.configFilePath,'r')
         try:
            eof = False
            while not eof:
               record = fpConfig.readline()
               if record != "":
                  record = commentPattern.sub("",record)
                  if   sectionPattern.match(record):
                     sectionName = sectionPattern.match(record).group(2)
                     inClientSection = (sectionName == 'client')
                     if inClientSection:
                        self.configData = {'listenURIs':[],
                                           'maximumConnectionPasses':15,
                                           'submitSSLCA':"/etc/submit/submit_server_ca.crt",
                                           'pegasusVersion':"4.1.0",
                                           'useSetup':"/etc/environ.sh",
                                           'doubleDashTerminator':False
                                          }
                  elif inClientSection:
                     if keyValuePattern.match(record):
                        key,value = keyValuePattern.match(record).group(2,4)
                        if key in self.configData:
                           if   isinstance(self.configData[key],list):
                              self.configData[key] = [e.strip() for e in value.split(',')]
                           elif isinstance(self.configData[key],bool):
                              self.configData[key] = bool(value.lower() == 'true')
                           elif isinstance(self.configData[key],float):
                              self.configData[key] = float(value)
                           elif isinstance(self.configData[key],int):
                              self.configData[key] = int(value)
                           elif isinstance(self.configData[key],dict):
                              try:
                                 sampleKey   = self.configData[key].keys()[0]
                                 sampleValue = self.configData[key][sampleKey]
                              except:
                                 sampleKey   = "key"
                                 sampleValue = "value"
                              self.configData[key] = {} 
                              for e in value.split(','):
                                 dictKey,dictValue = e.split(':')
                                 if isinstance(sampleKey,int):
                                    dictKey = int(dictKey)
                                 if   isinstance(sampleValue,int):
                                    dictValue = int(dictValue)
                                 elif isinstance(sampleValue,float):
                                    dictValue = float(dictValue)
                                 elif isinstance(sampleValue,bool):
                                    dictValue = bool(dictValue.lower() == 'true')
                                 self.configData[key][dictKey] = dictValue
                           else:
                              self.configData[key] = value
                        else:
                           self.logger.log(logging.WARNING,getLogMessage("Undefined key = value pair %s = %s" % (key,value)))
               else:
                  eof = True
         except (IOError,OSError):
            self.logger.log(logging.ERROR,getLogMessage("%s could not be read" % (self.configFilePath)))
         finally:
            fpConfig.close()
      except (IOError,OSError):
         self.logger.log(logging.ERROR,getLogMessage("%s could not be opened" % (self.configFilePath)))

      configured = True
      if len(self.configData['listenURIs']) == 0:
         self.logger.log(logging.ERROR,getLogMessage("listenURIs missing from %s" % (self.configFilePath)))
         configured = False

      self.setupPegasus()

      if 'SUBMIT_DOUBLEDASH' in os.environ:
         if   os.environ['SUBMIT_DOUBLEDASH'].lower() == 'true':
            self.configData['doubleDashTerminator'] = True
         elif os.environ['SUBMIT_DOUBLEDASH'].lower() == 'false':
            self.configData['doubleDashTerminator'] = False

      return(configured)


   def exit(self,
            exitCode=0):
      if self.clientData['readyToExit']:
         if not self.jobScanner:
            sys.exit(self.clientData['exitCode'])
      else:
         if self.serverConnection and self.serverConnection.isConnected():
            clientMessage = {'messageType':'clientExit',
                             'exitCode':exitCode}
            self.serverConnection.postJsonMessage(clientMessage)
            self.clientData['readyToExit'] = True
         else:
            self.clientData['exitCode'] = exitCode
            sys.exit(exitCode)


   def getUserAttributes(self):
      return(self.clientIdAuthAttributes.getUserAttributes())


   def haveIdAuthAttributes(self):
      return(self.clientIdAuthAttributes.haveIdAuthAttributes())


   def connect(self):
      isConnected = False
      if self.haveIdAuthAttributes():
         self.serverConnection = UnboundConnection(UnboundConnection.TLSREQUIREMENTNONE,
                                                   listenURIs=self.configData['listenURIs'],
                                                   maximumConnectionPasses=self.configData['maximumConnectionPasses'],
                                                   submitSSLCA=self.configData['submitSSLCA'])
         isConnected = self.serverConnection.isConnected()

      return(isConnected)


   def sendContext(self):
      clientMessage = {'messageType':'clientVersion',
                       'version':self.clientData['version']}
      self.serverConnection.postJsonMessage(clientMessage)
      clientMessage = {'messageType':'doubleDashTerminator',
                       'doubleDashTerminator':self.configData['doubleDashTerminator']}
      self.serverConnection.postJsonMessage(clientMessage)
      for arg in sys.argv[0:]:
         clientMessage = {'messageType':'arg',
                          'arg':arg}
         self.serverConnection.postJsonMessage(clientMessage)

      blackListEnvironmentVariables = ['DISPLAY' 'EDITOR','HOME','LOGNAME','LS_COLORS','MAIL', \
                                       'OLDPWD','RESULTSDIR','SESSIONDIR','SHELL','SHLVL', \
                                       'SVN_EDITOR','SUBMIT_DOUBLEDASH','TERM','TIMEOUT','USER','VISUAL', \
                                       'WINDOWID', 'XTERM_LOCALE','XTERM_SHELL','XTERM_VERSION','_']
      for environmentVar in os.environ:
         if not environmentVar in blackListEnvironmentVariables:
            clientMessage = {'messageType':'var',
                             'name':environmentVar,
                             'value':os.environ[environmentVar]}
            self.serverConnection.postJsonMessage(clientMessage)

      umask = os.umask(0)
      os.umask(umask)
      clientMessage = {'messageType':'umask',
                       'umask':umask}
      self.serverConnection.postJsonMessage(clientMessage)
      if self.clientData['workDirectoryDevice']:
         clientMessage = {'messageType':'pwdWithMount',
                          'path':self.clientData['workDirectory'],
                          'inode':self.clientData['workDirectoryInode'],
                          'device':self.clientData['workDirectoryDevice']}
         self.serverConnection.postJsonMessage(clientMessage)
      else:
         clientMessage = {'messageType':'pwd',
                          'path':self.clientData['workDirectory'],
                          'inode':self.clientData['workDirectoryInode']}
         self.serverConnection.postJsonMessage(clientMessage)
#     self.clientData['isClientTTY'] = sys.stdin.isatty() and sys.stdout.isatty()
      clientMessage = {'messageType':'isClientTTY',
                       'isClientTTY':self.clientData['isClientTTY']}
      self.serverConnection.postJsonMessage(clientMessage)
      if self.clientData['pegasusVersion']:
         clientMessage = {'messageType':'pegasusVersion',
                          'version':self.clientData['pegasusVersion']}
         self.serverConnection.postJsonMessage(clientMessage)
      else:
         clientMessage = {'messageType':'pegasusVersion',
                          'version':self.configData['pegasusVersion']}
         self.serverConnection.postJsonMessage(clientMessage)


   def signon(self):
      signonAttributes = self.clientIdAuthAttributes.getSignonAttributes()
      clientMessage = {'messageType':'userName',
                       'userName':signonAttributes['userName']}
      self.serverConnection.postJsonMessage(clientMessage)
      if 'password' in signonAttributes:
         clientMessage = {'messageType':'password',
                          'password':signonAttributes['password']}
         self.serverConnection.postJsonMessage(clientMessage)
      if 'sessionId' in signonAttributes:
         clientMessage = {'messageType':'sessionId',
                          'sessionId':int(signonAttributes['sessionId'])}
         self.serverConnection.postJsonMessage(clientMessage)
      if 'token' in signonAttributes:
         clientMessage = {'messageType':'sessionToken',
                          'token':signonAttributes['token']}
         self.serverConnection.postJsonMessage(clientMessage)
      clientMessage = {'messageType':'signon'}
      self.serverConnection.postJsonMessage(clientMessage)

      self.clientData['userName'] = signonAttributes['userName']
      if 'userId' in signonAttributes:
         self.clientData['userId'] = signonAttributes['userId']


   def mapSubmitCommandFiles(self):
      exitCode = 0
      if self.clientData['operationMode'] & self.commandParser.OPERATIONMODERUNDISTRIBUTORID:
         submitCommandFiles = self.commandParser.getSubmitCommandFiles()
         if len(submitCommandFiles) > 0:
            if len(self.clientData['mountPoints']) > 0:
               for submitCommandFile in submitCommandFiles:
                  submitCommandPath = os.path.abspath(submitCommandFile)
                  mountPoint = submitCommandPath
                  while mountPoint:
                     if os.path.ismount(mountPoint):
                        try:
                           inode = os.lstat(submitCommandFile).st_ino
                           fileInfo = {'argValue':submitCommandFile,
                                       'path':submitCommandPath,
                                       'inode':inode}
                           self.clientData['mountPoints'][mountPoint]['submitCommandFiles'].append(fileInfo)
                        except OSError,err:
                           self.__writeToStderr("%s: %s\n" % (submitCommandFile,err.args[1]))
                           exitCode = err.args[0]
                        break
                     else:
                        if mountPoint != os.sep:
                           mountPoint = os.path.dirname(mountPoint)
                        else:
                           break

               markedForDeletion = []
               for mountPoint in self.clientData['mountPoints']:
                  if len(self.clientData['mountPoints'][mountPoint]['submitCommandFiles']) == 0:
                     markedForDeletion.append(mountPoint)

               for mountPoint in markedForDeletion:
                  del self.clientData['mountPoints'][mountPoint]
               del markedForDeletion

               self.logger.log(logging.INFO,getLogMessage(self.clientData['mountPoints']))

               for mountPoint in self.clientData['mountPoints']:
                  device = self.clientData['mountPoints'][mountPoint]['device']
                  for fileInfo in self.clientData['mountPoints'][mountPoint]['submitCommandFiles']:
                     clientMessage = {'messageType':'submitCommandFileWithMount',
                                      'path':fileInfo['path'],
                                      'inode':fileInfo['inode'],
                                      'argValue':fileInfo['argValue'],
                                      'device':device}
                     self.serverConnection.postJsonMessage(clientMessage)
                     self.logger.log(logging.DEBUG,getLogMessage("submitCommandFileWithMount = %s" % (fileInfo['path'])))
            else:
               for submitCommandFile in submitCommandFiles:
                  try:
                     inode = os.lstat(submitCommandFile).st_ino
                     clientMessage = {'messageType':'submitCommandFile',
                                      'path':os.path.abspath(submitCommandFile),
                                      'argValue':submitCommandFile,
                                      'inode':inode}
                     self.serverConnection.postJsonMessage(clientMessage)
                  except OSError,err:
                     self.__writeToStderr("%s: %s\n" % (submitCommandFile,err.args[1]))
                     exitCode = err.args[0]

      if exitCode:
         self.exit(exitCode)
      else:
         clientMessage = {'messageType':'submitCommandFileInodesSent'}
         self.serverConnection.postJsonMessage(clientMessage)


   def connectToServer(self):
      connectedToServer = False
      if self.haveIdAuthAttributes():
         if self.connect():
            self.sendContext()
            self.clientData['authorizedAction'] = self.AUTHORIZEDACTIONEXECUTE
            self.signon()
            connectedToServer = True
         else:
            self.sendUserErrorNotification("Connection to submit server failed.")
            self.exit(1)
      else:
         self.sendUserErrorNotification("Could not acquire proper authentication information.")
         sys.exit(1)

      return(connectedToServer)


   def getInputObjects(self):
      serverReader = self.serverConnection.getInputObject()
      childProcessInputObjects = []
      if self.clientData['childReadStdout']:
         childProcessInputObjects.append(self.clientData['childReadStdout'])
      if self.clientData['childReadStderr']:
         childProcessInputObjects.append(self.clientData['childReadStderr'])

      return(serverReader,
             self.clientData['inputObjects'].keys(),
             self.clientData['exportObjects'].keys(),
             childProcessInputObjects)


   def getOutputObjects(self):
      serverWriter = self.serverConnection.getOutputObject()

      importObjects = []
      for importObject in self.clientData['importObjects']:
         if self.clientData['importObjects'][importObject]['buffer']:
            importObjects.append(importObject)

      return(serverWriter,importObjects)


   def readFile(self,
                fileObject):
      if   fileObject in self.clientData['inputObjects']:
         inputFile = self.clientData['inputObjects'][fileObject]
         try:
            text = os.read(fileObject.fileno(),1024)
            if text == "":
               clientMessage = {'messageType':'close',
                                'path':inputFile}
               self.serverConnection.postJsonMessage(clientMessage)
               self.closeFile(fileObject)
            else:
               clientMessage = {'messageType':'read',
                                'path':inputFile,
                                'text':text}
               self.serverConnection.postJsonMessage(clientMessage)
         except:
            self.exit(1)
      elif fileObject in self.clientData['exportObjects']:
         inputFile = self.clientData['exportObjects'][fileObject]
         try:
            text = os.read(fileObject.fileno(),1024)
            if text == "":
               clientMessage = {'messageType':'close',
                                'path':inputFile}
               self.serverConnection.postJsonMessage(clientMessage)
               self.closeFile(fileObject)
            else:
               clientMessage = {'messageType':'read',
                                'path':inputFile,
                                'text':text}
               self.serverConnection.postJsonMessage(clientMessage)
         except:
            self.exit(1)


   def writeFile(self,
                 fileObject):
      if fileObject in self.clientData['importObjects']:
         writeBuffer = self.clientData['importObjects'][fileObject]['buffer']
         try:
            textLength = os.write(fileObject.fileno(),writeBuffer)
            self.clientData['importObjects'][fileObject]['buffer'] = writeBuffer[textLength:]
         except:
            self.exit(1)


   def closeFile(self,
                 fileObject):
      if fileObject in self.clientData['inputObjects']:
         del self.clientData['inputObjects'][fileObject]

      if fileObject in self.clientData['exportObjects']:
         exportFile = self.clientData['exportObjects'][fileObject]
         exportPath = os.path.join(os.sep,"tmp",exportFile)
         if os.path.exists(exportPath):
            os.remove(exportPath)
         del self.clientData['exportObjects'][fileObject]

      if fileObject in self.clientData['outputObjects']:
         del self.clientData['outputObjects'][fileObject]

      transferTarFile = ""
      if fileObject in self.clientData['importObjects']:
         transferTarFile = self.clientData['importObjects'][fileObject]['path']
         del self.clientData['importObjects'][fileObject]
         if len(self.clientData['importObjects']) == 0:
            self.clientData['importFilesComplete'] = True
            clientMessage = {'messageType':'importFilesComplete'}
            self.serverConnection.postJsonMessage(clientMessage)

      if fileObject == self.clientData['childReadStdout']:
         self.clientData['childReadStdout'] = None
      if fileObject == self.clientData['childReadStderr']:
         self.clientData['childReadStderr'] = None

      try:
         fileObject.close()
         if transferTarFile:
            tarCommand = ['tar','xvf',os.path.join(self.clientData['workDirectory'],transferTarFile)]
            self.logger.log(logging.INFO,getLogMessage("command: %s" % (tarCommand)))
            child = subprocess.Popen(tarCommand,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE,
                                     close_fds=True)
            tarStdOutput,tarStdError = child.communicate()
            tarExitStatus = child.returncode
            if tarExitStatus == 0:
               os.remove(os.path.join(self.clientData['workDirectory'],transferTarFile))
            else:
               self.logger.log(logging.ERROR,getLogMessage("Failed to extract transfer tarfile %s" % (transferTarFile)))
               if tarStdOutput:
                  self.logger.log(logging.ERROR,getLogMessage(tarStdOutput))
               if tarStdError:
                  self.logger.log(logging.ERROR,getLogMessage(tarStdError))
      except:
         pass


   def checkClosePendingObjects(self):
      markedForDeletion = []
      for fileObject in self.clientData['closePendingObjects']:
         if fileObject in self.clientData['importObjects']:
            if self.clientData['importObjects'][fileObject]['buffer'] == "":
               self.closeFile(fileObject)
               markedForDeletion.append(fileObject)
      for fileObject in markedForDeletion:
         self.clientData['closePendingObjects'].remove(fileObject)
      del markedForDeletion


   def showHelpUsage(self,
                     operationMode):
      if operationMode & (self.commandParser.OPERATIONMODEHELPUSAGE | \
                          self.commandParser.OPERATIONMODEHELPEXAMPLES):
         if operationMode & (self.commandParser.OPERATIONMODEHELPTOOLS | \
                             self.commandParser.OPERATIONMODEHELPVENUES | \
                             self.commandParser.OPERATIONMODEHELPMANAGERS | \
                             self.commandParser.OPERATIONMODEVERSIONSERVER | \
                             self.commandParser.OPERATIONMODEVERSIONDISTRIBUTOR):
            showHelp = False
         else:
            showHelp = True
      else:
         showHelp = False

      return(showHelp)


   def parseCommandArguments(self):
      parseCommandError = False
      contactServer = False
      self.commandParser = CommandParser(self.configData['doubleDashTerminator'])
      self.commandParser.parseArguments(sys.argv[1:])
      self.clientData['operationMode'] = self.commandParser.getOperationMode()

      if self.commandParser.validateArguments():
         if self.clientData['operationMode'] & self.commandParser.OPERATIONMODEVERSIONCLIENT:
            self.__writeToStdout("Submit client version: %s\n" % (self.clientData['version']))

         if   self.showHelpUsage(self.clientData['operationMode']):
            self.commandParser.showUsage()
         elif self.clientData['operationMode'] & (self.commandParser.OPERATIONMODERUNSTATUS | \
                                                  self.commandParser.OPERATIONMODERUNKILL | \
                                                  self.commandParser.OPERATIONMODERUNVENUESTATUS | \
                                                  self.commandParser.OPERATIONMODERUNSERVER | \
                                                  self.commandParser.OPERATIONMODERUNDISTRIBUTOR):
            contactServer = True
         elif self.clientData['operationMode'] & self.commandParser.OPERATIONMODERUNPROXY:
            self.clientData['attachId'] = self.commandParser.getOption('attachId')
            contactServer = True
         elif self.clientData['operationMode'] & self.commandParser.OPERATIONMODERUNDISTRIBUTORID:
            self.commandParser.setSubmitCommandFiles()
            exitCode = self.commandParser.setSweepParameters()
            if exitCode == 0:
               exitCode = self.commandParser.setCSVDataParameters()
               if exitCode == 0:
                  parameterCombinationCount = self.commandParser.getParameterCombinationCount()
                  if parameterCombinationCount > 0:
                     self.clientData['isParametric'] = True

                  if not self.clientData['isParametric']:
                     if self.commandParser.getOption('progressCurses') or self.commandParser.getOption('progressSubmit'):
                        message = "submit and curses progress reports are only valid with parametric simulation\n"
                        self.__writeToStderr(message)
                        parseCommandError = True
                  else:
                     if self.commandParser.getOption('progressCurses') and not self.clientData['isClientTTY']:
                        message = "curses progress reports require a tty\n"
                        self.__writeToStderr(message)
                        parseCommandError = True

                  localExecution = self.commandParser.getOption('localExecution')
                  enteredExecutable = self.commandParser.getEnteredExecutable()
                  if   self.clientData['isParametric'] and not self.clientData['pegasusExists'] and localExecution:
                     message = "Pegasus is required to execute a parametric sweep.\n"
                     self.__writeToStderr(message)
                     if self.clientData['pegasusSetupError']:
                        self.__writeToStderr(self.clientData['pegasusSetupError'])
                     parseCommandError = True
                  elif self.clientData['isParametric'] and '@:' in enteredExecutable:
                     enteredExecutable = enteredExecutable.replace('@:','')
                     message = "Parameter substitution is not allowed within the command file: %s\n" % (enteredExecutable)
                     self.__writeToStderr(message)
                     parseCommandError = True
                  elif self.clientData['isParametric'] and '@@' in enteredExecutable:
                     message = "Parameter substitution is not allowed in the command name: %s\n" % (enteredExecutable)
                     self.__writeToStderr(message)
                     parseCommandError = True
                  elif self.clientData['isParametric'] and \
                       self.commandParser.getParametricArgumentCount() == 0 and \
                       self.commandParser.getParametricInputCount() == 0:
                     message = "Parameters supplied for %d jobs, " % (parameterCombinationCount) + \
                               "but there are no template files or\ncommand substitutions that would make the jobs unique.  " + \
                               "You should prefix any template files\nwith '@:' or " + \
                               "embed parameter names (in the form '@@name') into command line arguments.\n"
                     self.__writeToStderr(message)
                     parseCommandError = True
                  elif not self.checkInputsExist():
                     parseCommandError = True
                  else:
                     self.clientData['localExecution'] = localExecution
                     self.clientData['isLocalPegasus'] = os.path.basename(enteredExecutable).startswith('pegasus-')
                     self.clientData['reportMetrics']  = self.commandParser.getOption('metrics')
                     self.clientData['detach']         = self.commandParser.getOption('detach')
                     if self.commandParser.getOption('runName'):
                        self.clientData['runName']     = self.commandParser.getOption('runName')
                     contactServer = True

                     if self.clientData['localExecution']:
                        if self.clientData['isParametric'] or self.clientData['isLocalPegasus']:
                           if self.clientData['runName']:
                              runDirectory = os.path.join(os.getcwd(),self.clientData['runName'])
                              if os.path.exists(runDirectory):
                                 message = "Run directory %s exists\n" % (runDirectory)
                                 self.__writeToStderr(message)
                                 parseCommandError = True
                                 contactServer     = False
               else:
                  parseCommandError = True
            else:
               parseCommandError = True
         elif self.clientData['operationMode'] & self.commandParser.OPERATIONMODERUNNONE:
            if not self.clientData['operationMode'] & self.commandParser.OPERATIONMODEVERSIONCLIENT:
               if not self.clientData['operationMode'] & self.commandParser.OPERATIONMODEHELPEXAMPLES:
                  message = "No command given to execute\n"
                  self.__writeToStderr(message)
                  parseCommandError = True
      else:
         parseCommandError = True

      if   not self.commandParser.getOption('progress'):
         self.clientData['progressReport'] = 'silent'
      elif self.commandParser.getOption('progressCurses'):
         self.clientData['progressReport'] = 'curses'
      elif self.commandParser.getOption('progressSubmit'):
         self.clientData['progressReport'] = 'submit'
      elif self.commandParser.getOption('progressText'):
         self.clientData['progressReport'] = 'text'
      else:
         if self.clientData['isParametric'] and self.clientData['isClientTTY']:
            self.clientData['progressReport'] = 'curses'
         else:
            self.clientData['progressReport'] = 'text'

      return(parseCommandError,contactServer)


   def reportDebugOutput(self):
      return(self.commandParser.getOption('debug'))


   def __updatePegasusWorkflowStatus(self):
      jobStatusReportOrders = {}
      jobStatusReportOrders['waiting']    = 5
      jobStatusReportOrders['aborted']    = 1
      jobStatusReportOrders['setup']      = 6
      jobStatusReportOrders['setting up'] = 6
      jobStatusReportOrders['failed']     = 3
      jobStatusReportOrders['executing']  = 4
      jobStatusReportOrders['finished']   = 2

      parameterCombinationsDir  = os.path.dirname(self.clientData['jobScanPath'])
      parameterCombinationsBase = os.path.basename(self.clientData['jobScanPath'])
      if '.' in parameterCombinationsBase:
         parameterCombinationsBase = parameterCombinationsBase.split('.')[0]
      tmpParameterCombinationsFile = parameterCombinationsBase + '.tmp'
      tmpParameterCombinationsPath = os.path.join(parameterCombinationsDir,tmpParameterCombinationsFile)
      copyTmpFile = False

      if os.path.exists(self.clientData['jobScanPath']):
         try:
            fpCSVIn = open(self.clientData['jobScanPath'],'rb')
            try:
               csvReader = csv.reader(fpCSVIn)
               try:
                  fpCSVOut = open(tmpParameterCombinationsPath,'wb')
                  try:
                     csvWriter = csv.writer(fpCSVOut)
                     csvWriter.writerow(('# command: ' + self.clientData['enteredCommand'],))
                     csvWriter.writerow(('# started: ' + self.clientData['startDate'],))
                     if self.clientData['finishDate']:
                        csvWriter.writerow(('# finished: ' + self.clientData['finishDate'],))
                     nCompleted = 0
                     for instance in self.clientData['jobStates']:
                        jobStatusState = self.clientData['jobStates'][instance]
                        if jobStatusState in ['finished','failed','aborted']:
                           nCompleted += 1
                     csvWriter.writerow(('# completed: %d/%d jobs' % (nCompleted,len(self.clientData['jobStates'])),))
                  except csv.Error,e:
                     self.logger.log(logging.ERROR,getLogMessage("csv writter failed on %s" % (tmpParameterCombinationsPath)))
                  else:
                     parameterNames = []
                     try:
                        while len(parameterNames) <= 1:
                           parameterNames = csvReader.next()
                     except StopIteration:
                        self.logger.log(logging.ERROR,getLogMessage("%s header row is missing" % (self.clientData['jobScanPath'])))
                     except csv.Error,e:
                        self.logger.log(logging.ERROR,getLogMessage("csv reader failed on %s" % (self.clientData['jobScanPath'])))
                     else:
                        try:
                           csvWriter.writerow(parameterNames)
                        except csv.Error,e:
                           self.logger.log(logging.ERROR,getLogMessage("csv writter failed on %s" % (tmpParameterCombinationsPath)))
                        else:
                           try:
                              parameterCombinations = {}
                              for parameterCombination in csvReader:
                                 instance = int(parameterCombination[0])
                                 if instance in self.clientData['jobStates']:
                                    jobStatusState       = self.clientData['jobStates'][instance]
                                    jobStatusReportOrder = jobStatusReportOrders[jobStatusState]
                                    parameterCombination[1] = jobStatusState
                                    if not jobStatusReportOrder in parameterCombinations:
                                       parameterCombinations[jobStatusReportOrder] = []
                                    parameterCombinations[jobStatusReportOrder].append(parameterCombination)
                              activeJobStatusReportOrders = parameterCombinations.keys()
                              activeJobStatusReportOrders.sort()
                           except csv.Error,e:
                              self.logger.log(logging.ERROR,getLogMessage("csv reader failed on %s" % \
                                                                          (self.clientData['jobScanPath'])))
                           else:
                              try:
                                 for jobStatusReportOrder in activeJobStatusReportOrders:
                                    for parameterCombination in parameterCombinations[jobStatusReportOrder]:
                                       csvWriter.writerow(parameterCombination)
                              except csv.Error,e:
                                 self.logger.log(logging.ERROR,getLogMessage("csv writter failed on %s" % \
                                                                             (tmpParameterCombinationsPath)))
                              else:
                                 copyTmpFile = True
                  finally:
                     fpCSVOut.close()
               except (IOError,OSError):
                  self.logger.log(logging.ERROR,getLogMessage("%s could not be opened" % (tmpParameterCombinationsPath)))
            except csv.Error,e:
               self.logger.log(logging.ERROR,getLogMessage("csv reader failed on %s" % (self.clientData['jobScanPath'])))
            finally:
               fpCSVIn.close()
         except (IOError,OSError):
            self.logger.log(logging.ERROR,getLogMessage("%s could not be opened" % (self.clientData['jobScanPath'])))

      if copyTmpFile:
         os.rename(tmpParameterCombinationsPath,self.clientData['jobScanPath'])

      jobStatusStates = {}
      for jobStatusState in jobStatusReportOrders:
         jobStatusStates[jobStatusState] = 0
      nCompleted = 0
      for instance in self.clientData['jobStates']:
         jobStatusState = self.clientData['jobStates'][instance]
         if not jobStatusState in jobStatusStates:
            jobStatusStates[jobStatusState] = 0
         jobStatusStates[jobStatusState] += 1
         if jobStatusState in ['finished','failed','aborted']:
            nCompleted += 1
      percentDone = 100.*float(nCompleted)/float(len(self.clientData['jobStates']))

      jobStatusReports = {}
      for jobStatusState in jobStatusStates:
         jobStatusReportOrder = jobStatusReportOrders[jobStatusState]
         if not jobStatusReportOrder in jobStatusReports:
            jobStatusReports[jobStatusReportOrder] = {}
         jobStatusReports[jobStatusReportOrder][jobStatusState] = jobStatusStates[jobStatusState]

      activeJobStatusReportOrders = jobStatusReports.keys()
      activeJobStatusReportOrders.sort()
      runProgressMessage = "=SUBMIT-PROGRESS=>"
      for jobStatusReportOrder in activeJobStatusReportOrders:
         for jobStatusState in jobStatusReports[jobStatusReportOrder]:
            runProgressMessage += " %s=%d" % (jobStatusState.replace(' ','_'), \
                                              jobStatusReports[jobStatusReportOrder][jobStatusState])
      runProgressMessage += " %%done=%.2f" % (percentDone)
      if self.clientData['progressReport'] == 'submit':
         if runProgressMessage != self.clientData['runProgressMessage']:
            self.__writeToStdout("%s timestamp=%.1f\n" % (runProgressMessage,time.time()))
      self.clientData['runProgressMessage'] = runProgressMessage


   @staticmethod
   def __which(program):
      def isExe(fpath):
         return(os.path.isfile(fpath) and os.access(fpath,os.X_OK))

      fpath,fname = os.path.split(program)
      if fpath:
         if isExe(program):
            return(program)
      else:
         for path in os.environ["PATH"].split(os.pathsep):
            exePath = os.path.join(path,program)
            if isExe(exePath):
               return(exePath)

      return(None)


   def __getStdInputFile(self):
      stdInputFile = '/dev/null'
      if not sys.stdin.isatty():
         toCheck = []
         stdinFd = sys.stdin.fileno()
         toCheck.append(stdinFd)
         try:
            ready = select.select(toCheck,[],[],0.) # wait for input
         except select.error,err:
            ready = {}
            ready[0] = []
         if stdinFd in ready[0]:
            content = sys.stdin.read()
            if content != "":
               stdInputFile = os.path.join(self.clientData['jobPath'],".__%s.stdin" % (self.clientData['localJobId']))
               try:
                  stdinFile = open(stdInputFile,'w')
                  try:
                     stdinFile.write(content)
                  except (IOError,OSError):
                     self.logger.log(logging.ERROR,getLogMessage("%s could not be written" % (stdInputFile)))
                  finally:
                     stdinFile.close()
               except (IOError,OSError):
                  self.logger.log(logging.ERROR,getLogMessage("%s could not be opened" % (stdInputFile)))
               try:
                  si = open('/dev/tty')
               except:
                  pass
               else:
                  os.dup2(si.fileno(),sys.stdin.fileno())
            del content

      return(stdInputFile)


   def getUseEnvironment(self):
      useEnvironment = ""
      reChoice = re.compile(".*_CHOICE$")
      environmentVars = filter(reChoice.search,os.environ.keys())
      if "PROMPT_CHOICE" in environmentVars:
         environmentVars.remove("PROMPT_CHOICE")
      if "EDITOR_CHOICE" in environmentVars:
         environmentVars.remove("EDITOR_CHOICE")
      if len(environmentVars) > 0:
         includedScripts = []
         useEnvironment  = "ENVIRON_CONFIG_DIRS=\"\"\n"
         useEnvironment += ". %s\n" % (self.configData['useSetup'])
         if "ENVIRON_CONFIG_DIRS" in os.environ:
            useEnvironment += "export ENVIRON_CONFIG_DIRS=\"" + os.environ['ENVIRON_CONFIG_DIRS'] + "\"\n"
         for environmentVar in environmentVars:
            includedScript = os.environ[environmentVar]
            if not includedScript in includedScripts:
               useEnvironment += "use -e -r " + includedScript + "\n"
               includedScripts.append(includedScript)
         del includedScripts

      return(useEnvironment)


   def startLocalParametric(self):
      from LocalWorkflowPEGASUS import LocalWorkflowPEGASUS
      exitCode = 0

      useEnvironment = self.getUseEnvironment()

      environment = ""
      if self.commandParser.getOption('environment'):
         for environmentVariableValue in self.commandParser.getOption('environment'):
            environmentVariable = ""
            value               = ""
            nParts = len(environmentVariableValue.split('='))
            if   nParts == 1:
               environmentVariable = environmentVariableValue.strip()
               if environmentVariable in os.environ:
                  value = os.environ[environmentVariable]
            elif nParts == 2:
               environmentVariable,value = environmentVariableValue.split('=')
               environmentVariable = environmentVariable.strip()
               value               = value.strip()
               if value == "":
                  if environmentVariable in os.environ:
                     value = os.environ[environmentVariable]
            if environmentVariable == "" or value == "":
               message = "Invalid environment variable %s specified." % (environmentVariableValue)
               self.logger.log(logging.ERROR,getLogMessage(message))
               if not exitCode:
                  exitCode = 1
            else:
               environment += environmentVariable + "=" + value + " "
         environment = environment.strip()

      clientInputFiles = []
      if self.commandParser.getOption('inputFiles'):
         for inputFile in self.commandParser.getOption('inputFiles'):
            clientInputFiles.append(inputFile)

      self.clientData['enteredCommand'] = self.commandParser.getEnteredCommand()
      enteredCommandArguments = self.commandParser.getEnteredCommandArguments()

      self.clientData['jobPath'] = os.path.join(os.getcwd(),self.clientData['runName'])
      if not os.path.isdir(self.clientData['jobPath']):
         os.mkdir(self.clientData['jobPath'])
      self.clientData['jobScanPath'] = os.path.join(self.clientData['jobPath'],'parameterCombinations.csv')
      self.commandParser.writeParameterCombinations(self.clientData['jobScanPath'])

      self.stdinput = self.__getStdInputFile()

      inputError = False
      executableExistsError = False
      nInstanceIdDigits = max(2,int(math.log10(self.commandParser.getParameterCombinationCount())+1))
      instance = 0
      for substitutions in self.commandParser.getNextParameterCombinationFromCSV(self.clientData['jobScanPath']):
         instance += 1
         instanceId = str(instance).zfill(nInstanceIdDigits)
         self.clientData['jobStates'][instance] = 'setup'

         instanceDirectory = os.path.join(self.clientData['jobPath'],instanceId)
         if not os.path.isdir(instanceDirectory):
            os.makedirs(instanceDirectory)
         instanceInputsPath = instanceDirectory

         exitCode = 0

         enteredCommand = self.commandParser.getEnteredCommand()
         template = ParameterTemplate(enteredCommand)
         try:
            enteredCommand = template.substitute(substitutions)
         except KeyError,err:
            self.__writeToStderr("Pattern substitution failed for @@%s\n" % (err[0]))
            inputError = True

         inputFiles = []
         for inputFile in clientInputFiles:
            template = ParameterTemplate(inputFile)
            try:
               inputFile = template.substitute(substitutions)
            except KeyError,err:
               self.__writeToStderr("Pattern substitution failed for @@%s\n" % (err[0]))
               inputError = True
            inputFiles.append(inputFile)

         if len(enteredCommandArguments) > 1:
            for arg in enteredCommandArguments[1:]:
               if not arg.startswith('-'):
                  template = ParameterTemplate(arg)
                  try:
                     arg = template.substitute(substitutions)
                  except KeyError,err:
                     self.__writeToStderr("Pattern substitution failed for @@%s\n" % (err[0]))
                     inputError = True
                  if os.path.isfile(arg.replace('@:','')):
                     if not arg in inputFiles:
                        inputFiles.append(arg)

         substitutedInputFiles = []
         for inputFile in inputFiles:
            if '@:' in inputFile:
               inputFile = inputFile.replace('@:','')
               try:
                  fpInputFile = open(inputFile,'r')
                  try:
                     inputText = fpInputFile.readlines()
                  except (IOError,OSError):
                     self.logger.log(logging.ERROR,getLogMessage("%s could not be read" % (inputFile)))
                  else:
                     template = ParameterTemplate(''.join(inputText))
                     try:
                        inputText = template.substitute(substitutions)
                     except KeyError,err:
                        self.__writeToStderr("Pattern substitution failed for @@%s\n" % (err[0]))
                        inputError = True
                     inputName = os.path.basename(inputFile)
                     inputPath = os.path.join(instanceInputsPath,inputName)
                     try:
                        fpInputPath = open(inputPath,'w')
                        try:
                           fpInputPath.writelines(inputText)
                        except (IOError,OSError):
                           self.logger.log(logging.ERROR,getLogMessage("%s could not be written" % (inputPath)))
                        else:
                           substitutedInputFiles.append(inputPath)
                        finally:
                           fpInputPath.close()
                     except (IOError,OSError):
                        self.logger.log(logging.ERROR,getLogMessage("%s could not be opened" % (inputPath)))
                  finally:
                     fpInputFile.close()
               except (IOError,OSError):
                  self.logger.log(logging.ERROR,getLogMessage("%s could not be opened" % (inputFile)))
            else:
               substitutedInputFiles.append(inputFile)
         del inputFiles
         inputFiles = substitutedInputFiles

         remoteArgs = []

         executable = self.__which(enteredCommandArguments[0])
         if executable:
            executable = os.path.realpath(executable)
            remoteArgs.append(executable)
         else:
            if not executableExistsError:
               self.__writeToStderr("Specified command %s is not in your PATH\n" % (enteredCommandArguments[0]))
            inputError = True
            # allow completion of input checks
            remoteArgs.append(enteredCommandArguments[0])
            executable = enteredCommandArguments[0]
            executableExistsError = True

         for arg in enteredCommandArguments[1:]:
            if arg.startswith('-'):
               remoteArgs.append(arg)
            else:
               template = ParameterTemplate(arg)
               try:
                  arg = template.substitute(substitutions)
               except KeyError,err:
                  self.__writeToStderr("Pattern substitution failed for @@%s\n" % (err[0]))
                  inputError = True
               if '@:' in arg:
                  if os.path.isfile(arg.replace('@:','')):
                     inputName = os.path.basename(arg.replace('@:',''))
                     remoteArgs.append(inputName)
               else:
                  if os.path.isfile(arg):
                     rarg = os.path.abspath(arg)
                     remoteArgs.append(rarg)
                  else:
                     remoteArgs.append(arg)

         remoteCommand = " ".join(remoteArgs)
         arguments     = remoteArgs[1:]

         timeHistoryLogs = {}
         timeHistoryLogs['timestampTransferred'] = "%s.%s_%s" % (TIMESTAMPTRANSFERRED,self.clientData['localJobId'],instanceId)
         timeHistoryLogs['timestampStart']       = "%s.%s_%s" % (TIMESTAMPSTART,self.clientData['localJobId'],instanceId)
         timeHistoryLogs['timestampFinish']      = "%s.%s_%s" % (TIMESTAMPFINISH,self.clientData['localJobId'],instanceId)
         timeHistoryLogs['timeResults']          = "%s.%s_%s" % (TIMERESULTS,self.clientData['localJobId'],instanceId)
         timeHistoryLogs['exitCode']             = "%s.%s_%s" % (EXITCODE,self.clientData['localJobId'],instanceId)

         localBatchAppScript = LocalBatchAppScript(self.clientData['userName'],self.clientData['userId'],
                                                   self.serverData['submitterClass'],
                                                   self.clientData['runName'],instanceDirectory,
                                                   self.clientData['localJobId'],instanceId,
                                                   executable,self.stdinput,arguments,useEnvironment,environment,
                                                   self.serverData['submissionScripts'],timeHistoryLogs)
         appScriptName,appScript,appScriptExecutable = localBatchAppScript.buildAppScript()
         appScriptPath = os.path.join(self.clientData['jobPath'],instanceId,appScriptName)

         try:
            fpAppScript = open(appScriptPath,'w')
            try:
               fpAppScript.write(appScript)
            except (IOError,OSError):
               self.logger.log(logging.ERROR,getLogMessage("%s could not be written" % (appScriptPath)))
               exitCode = 1
            else:
               if appScriptExecutable:
                  os.chmod(appScriptPath,stat.S_IRWXU|stat.S_IRGRP|stat.S_IXGRP|stat.S_IROTH|stat.S_IXOTH)
               self.clientData['filesToRemove'].append(appScriptPath)
            finally:
               fpAppScript.close()
         except (IOError,OSError):
            self.logger.log(logging.ERROR,getLogMessage("%s could not be opened" % (appScriptPath)))
            exitCode = 1
         del localBatchAppScript

         stdFile = os.path.join(instanceDirectory,"%s_%s.stderr" % (self.clientData['runName'],instanceId))
         self.clientData['emptyFilesToRemove'].append(stdFile)

      nInstances = instance

      instance = 0
      instanceId = str(instance).zfill(nInstanceIdDigits)
      timeHistoryLogs = {}
      timeHistoryLogs['timestampTransferred'] = "%s.%s_%s" % (TIMESTAMPTRANSFERRED,self.clientData['localJobId'],instanceId)
      timeHistoryLogs['timestampStart']       = "%s.%s_%s" % (TIMESTAMPSTART,self.clientData['localJobId'],instanceId)
      timeHistoryLogs['timestampFinish']      = "%s.%s_%s" % (TIMESTAMPFINISH,self.clientData['localJobId'],instanceId)
      timeHistoryLogs['timeResults']          = "%s.%s_%s" % (TIMERESULTS,self.clientData['localJobId'],instanceId)

      pegasusTemplates = {}
      if self.serverData['pegasusRC'] != "":
         pegasusTemplates['rc'] = self.serverData['pegasusRC']
      if self.serverData['pegasusSites'] != "":
         pegasusTemplates['sites'] = self.serverData['pegasusSites']

      self.localWorkflowPEGASUS = LocalWorkflowPEGASUS(self.clientData['userName'],self.clientData['userId'],
                                                       self.serverData['submitterClass'],self.clientData['runName'],
                                                       self.clientData['jobPath'],self.clientData['jobPath'],
                                                       self.configData['useSetup'],
                                                       self.clientData['pegasusVersion'],self.clientData['pegasusHome'],
                                                       self.clientData['localJobId'],instanceId,
                                                       pegasusTemplates,self.serverData['submissionScripts'],timeHistoryLogs)
      batchScriptName,batchScript,batchScriptExecutable = self.localWorkflowPEGASUS.buildWorkflowScript()
      batchScriptPath = os.path.join(self.clientData['jobPath'],batchScriptName)

      try:
         fpBatchScript = open(batchScriptPath,'w')
         try:
            fpBatchScript.write(batchScript)
         except (IOError,OSError):
            self.logger.log(logging.ERROR,getLogMessage("%s could not be written" % (batchScriptPath)))
            exitCode = 1
         else:
            if batchScriptExecutable:
               os.chmod(batchScriptPath,stat.S_IRWXU|stat.S_IRGRP|stat.S_IXGRP|stat.S_IROTH|stat.S_IXOTH)
            self.clientData['filesToRemove'].append(batchScriptPath)
         finally:
            fpBatchScript.close()
      except (IOError,OSError):
         self.logger.log(logging.ERROR,getLogMessage("%s could not be opened" % (batchScriptPath)))
         exitCode = 1

      enteredCommand = self.commandParser.getEnteredCommand()

      stdFile = os.path.join(self.clientData['jobPath'],"%s.stdout" % (self.clientData['runName']))
      self.clientData['emptyFilesToRemove'].append(stdFile)
      stdFile = os.path.join(self.clientData['jobPath'],"%s.stderr" % (self.clientData['runName']))
      self.clientData['emptyFilesToRemove'].append(stdFile)

      self.clientData['startDate']      = time.strftime("%a %b %e %X %Z %Y")
      self.clientData['localStartTime'] = time.time()
      self.__updatePegasusWorkflowStatus()

      if inputError:
         self.exit(1)
      else:
         self.clientData['childHasExited'] = False
         child = subprocess.Popen(batchScriptPath,shell=False,bufsize=1,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  close_fds=True)
         self.clientData['childPid']        = child.pid
         self.clientData['childReadStdout'] = child.stdout
         self.clientData['childReadStderr'] = child.stderr

         clientMessage = {'messageType':'localWait'}
         self.serverConnection.postJsonMessage(clientMessage)

         if not sys.stdin in self.clientData['inputObjects']:
            self.clientData['inputObjects'][sys.stdin] = '#STDIN#'

#        self.clientData['isClientTTY'] = sys.stdin.isatty() and sys.stdout.isatty()
         if self.clientData['progressReport'] == 'curses':
            self.jobScanner = JobScanner(self.clientData['jobScanPath'])
            self.clientData['jobScannerStarted'] = True
            if not self.jobScanner.start():
               self.jobScanner.finish()
               self.jobScanner = None


   def startLocalPegasusCommand(self):
      from LocalBatchPEGASUS import LocalBatchPEGASUS
      exitCode = 0

      self.clientData['enteredCommand'] = self.commandParser.getEnteredCommand()
      enteredCommandArguments = self.commandParser.getEnteredCommandArguments()
      executable = enteredCommandArguments[0]
      remoteArgs = []
      for arg in enteredCommandArguments[1:]:
         if   arg.startswith('-'):
            remoteArgs.append(arg)
         elif os.path.isfile(arg) or os.path.isdir(arg):
            rarg = os.path.abspath(arg)
            remoteArgs.append(rarg)
         else:
            remoteArgs.append(arg)
      arguments = " ".join(remoteArgs)

      instance = 1
      nInstanceIdDigits = 2
      instanceId = str(instance).zfill(nInstanceIdDigits)
      self.clientData['jobPath'] = os.path.join(os.getcwd(),self.clientData['runName'],instanceId)
      if not os.path.isdir(self.clientData['jobPath']):
         os.makedirs(self.clientData['jobPath'])

      timeHistoryLogs = {}
      timeHistoryLogs['timestampTransferred'] = "%s.%s_%s" % (TIMESTAMPTRANSFERRED,self.clientData['localJobId'],instanceId)
      timeHistoryLogs['timestampStart']       = "%s.%s_%s" % (TIMESTAMPSTART,self.clientData['localJobId'],instanceId)
      timeHistoryLogs['timestampFinish']      = "%s.%s_%s" % (TIMESTAMPFINISH,self.clientData['localJobId'],instanceId)
      timeHistoryLogs['timeResults']          = "%s.%s_%s" % (TIMERESULTS,self.clientData['localJobId'],instanceId)

      pegasusTemplates = {}
      if self.serverData['pegasusRC'] != "":
         pegasusTemplates['rc'] = self.serverData['pegasusRC']
      if self.serverData['pegasusSites'] != "":
         pegasusTemplates['sites'] = self.serverData['pegasusSites']

      self.localBatchPEGASUS = LocalBatchPEGASUS(self.clientData['userName'],self.clientData['userId'],
                                                 self.serverData['submitterClass'],self.clientData['runName'],
                                                 self.clientData['localJobId'],instanceId,
                                                 self.clientData['jobPath'],self.clientData['jobPath'],
                                                 self.configData['useSetup'],
                                                 self.clientData['pegasusVersion'],self.clientData['pegasusHome'],
                                                 executable,arguments,
                                                 pegasusTemplates,self.serverData['submissionScripts'],timeHistoryLogs)

      batchLogName,batchScriptName,batchScript,batchScriptExecutable = self.localBatchPEGASUS.buildBatchScript()
      batchScriptPath = os.path.join(self.clientData['jobPath'],batchScriptName)

      try:
         fpBatchScript = open(batchScriptPath,'w')
         try:
            fpBatchScript.write(batchScript)
         except (IOError,OSError):
            self.logger.log(logging.ERROR,getLogMessage("%s could not be written" % (batchScriptPath)))
            exitCode = 1
         else:
            if batchScriptExecutable:
               os.chmod(batchScriptPath,stat.S_IRWXU|stat.S_IRGRP|stat.S_IXGRP|stat.S_IROTH|stat.S_IXOTH)
            self.clientData['filesToRemove'].append(batchScriptPath)
         finally:
            fpBatchScript.close()
      except (IOError,OSError):
         self.logger.log(logging.ERROR,getLogMessage("%s could not be opened" % (batchScriptPath)))
         exitCode = 1

      stdFile = os.path.join(self.clientData['jobPath'],"%s.stdout" % (self.clientData['runName']))
      self.clientData['emptyFilesToRemove'].append(stdFile)
      stdFile = os.path.join(self.clientData['jobPath'],"%s.stderr" % (self.clientData['runName']))
      self.clientData['emptyFilesToRemove'].append(stdFile)

      self.clientData['localStartTime'] = time.time()

      self.clientData['childHasExited'] = False
      child = subprocess.Popen(batchScriptPath,shell=False,bufsize=1,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               close_fds=True)
      self.clientData['childPid']        = child.pid
      self.clientData['childReadStdout'] = child.stdout
      self.clientData['childReadStderr'] = child.stderr

      clientMessage = {'messageType':'localWait'}
      self.serverConnection.postJsonMessage(clientMessage)


   def startLocalExecute(self):
      self.clientData['localStartTime'] = time.time()
      commandArguments = self.commandParser.getEnteredCommandArguments()
      os.environ['SUBMIT_JOB'] = "%s" % (self.serverData['jobId'])
      try:
         self.clientData['childHasExited'] = False
         child = subprocess.Popen(commandArguments,shell=False,bufsize=1,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  env=os.environ,
                                  close_fds=True)
         self.clientData['childPid']        = child.pid
         self.clientData['childReadStdout'] = child.stdout
         self.clientData['childReadStderr'] = child.stderr
         clientMessage = {'messageType':'localWait'}
         self.serverConnection.postJsonMessage(clientMessage)
         if self.serverData['heartbeatInterval'] > 0:
            self.scheduleHeartbeat()
      except OSError,err:
         self.__writeToStderr("Failed to invoke %s: %s\n" % (commandArguments[0],err.args[1]))
         self.exit(err.args[0])


   def startLocal(self):
      if   self.clientData['isParametric']:
         self.startLocalParametric()
      elif self.clientData['isLocalPegasus']:
         self.startLocalPegasusCommand()
      else:
         self.startLocalExecute()

      signal.signal(signal.SIGTSTP,signal.SIG_DFL)
      signal.signal(signal.SIGCONT,signal.SIG_DFL)


   def finishLocalParametric(self):
      if os.path.isfile(self.stdinput):
         os.remove(self.stdinput)

      parameterCombinationCount = self.commandParser.getParameterCombinationCount()
      if parameterCombinationCount > 0:
         nInstanceIdDigits = max(2,int(math.log10(parameterCombinationCount)+1))
         for instance in xrange(1,parameterCombinationCount+1):
            instanceId = str(instance).zfill(nInstanceIdDigits)

            jobStatistic = JobStatistic(1)
            timeFile = "%s.%s_%s" % (TIMESTAMPSTART,self.clientData['localJobId'],instanceId)
            timePath = os.path.join(self.clientData['jobPath'],instanceId,timeFile)
            jobStatistic.recordTime('jobStartedTime',timePath)
            timeFile = "%s.%s_%s" % (TIMESTAMPFINISH,self.clientData['localJobId'],instanceId)
            timePath = os.path.join(self.clientData['jobPath'],instanceId,timeFile)
            jobStatistic.recordTime('jobFinshedTime',timePath)
            timerFile = "%s.%s_%s" % (TIMERESULTS,self.clientData['localJobId'],instanceId)
            timerPath = os.path.join(self.clientData['jobPath'],instanceId,timerFile)
            jobStatistic.recordTimer(timerPath)
            jobStatistic.setWaitingTime()
            jobStatistic.setElapsedRunTime()

            exitCodeFile = "%s.%s_%s" % (EXITCODE,self.clientData['localJobId'],instanceId)
            exitCodePath = os.path.join(self.clientData['jobPath'],instanceId,exitCodeFile)
            if os.path.exists(exitCodePath):
               os.remove(exitCodePath)

            exitCode       = jobStatistic['exitCode']
            cpuTime        = jobStatistic['userTime']+jobStatistic['sysTime']
            elapsedRunTime = jobStatistic['elapsedRunTime']

            for ftype in '.dax','_sites.xml','.pegasusrc','_tc.txt':
               pegasusFile = os.path.join(self.clientData['jobPath'],"%s_%s%s" % (self.clientData['localJobId'],instanceId,ftype))
               if os.path.isfile(pegasusFile):
                  os.remove(pegasusFile)
               else:
                  pegasusFile = os.path.join(self.clientData['jobPath'],"%s%s" % (self.clientData['localJobId'],ftype))
                  if os.path.isfile(pegasusFile):
                     os.remove(pegasusFile)

            clientMessage = {'messageType':'localMetrics',
                             'instanceIndex':instance,
                             'status':exitCode,
                             'cpuTime':cpuTime,
                             'realTime':elapsedRunTime}
            self.serverConnection.postJsonMessage(clientMessage)
            del jobStatistic

         instance = 0
         instanceId = str(instance).zfill(nInstanceIdDigits)
         jobStatistic = JobStatistic(1)
         timeFile = "%s.%s_%s" % (TIMESTAMPSTART,self.clientData['localJobId'],instanceId)
         timePath = os.path.join(self.clientData['jobPath'],timeFile)
         jobStatistic.recordTime('jobStartedTime',timePath)
         timeFile = "%s.%s_%s" % (TIMESTAMPFINISH,self.clientData['localJobId'],instanceId)
         timePath = os.path.join(self.clientData['jobPath'],timeFile)
         jobStatistic.recordTime('jobFinshedTime',timePath)
         timerFile = "%s.%s_%s" % (TIMERESULTS,self.clientData['localJobId'],instanceId)
         timerPath = os.path.join(self.clientData['jobPath'],timerFile)
         jobStatistic.recordTimer(timerPath)
         jobStatistic.setWaitingTime()
         jobStatistic.setElapsedRunTime()

         exitCode       = jobStatistic['exitCode']
         cpuTime        = jobStatistic['userTime']+jobStatistic['sysTime']
         elapsedRunTime = jobStatistic['elapsedRunTime']

         clientMessage = {'messageType':'localMetrics',
                          'instanceIndex':instance,
                          'status':exitCode,
                          'cpuTime':cpuTime,
                          'realTime':elapsedRunTime}
         self.serverConnection.postJsonMessage(clientMessage)
         del jobStatistic

         workDirectoryName = os.path.join(self.clientData['jobPath'],'work')
         if os.path.isdir(workDirectoryName):
            shutil.rmtree(workDirectoryName,True)
         workDirectoryName = os.path.join(self.clientData['jobPath'],'scratch')
         if os.path.isdir(workDirectoryName):
            shutil.rmtree(workDirectoryName,True)

         if self.clientData['progressReport'] == 'text' or self.clientData['progressReport'] == 'pagasus':
            self.__writeToStdout("Simulations complete. Results are stored in directory %s\n" % (self.clientData['jobPath']))


   def moveTree(self,
                src,
                dst,
                symlinks=False):
      if os.path.isdir(src):
         if os.path.exists(dst):
            if not os.path.isdir(dst):
               self.logger.log(logging.ERROR,getLogMessage("moveTree: %s must be a directory" % (dst)))
               return
         else:
            os.mkdir(dst)
         names = os.listdir(src)
         for name in names:
            srcPath = os.path.join(src,name)
            dstPath = os.path.join(dst,name)
            try:
               if symlinks and os.path.islink(srcPath):
                  linkto = os.readlink(srcPath)
                  os.symlink(linkto,dstPath)
               elif os.path.isdir(srcPath):
                  self.moveTree(srcPath,dstPath,symlinks)
               else:
                  os.rename(srcPath,dstPath)
            except (IOError,os.error),why:
               self.logger.log(logging.ERROR,getLogMessage("moveTree: Can't move %s to %s: %s" % (srcPath,dstPath,str(why))))
      else:
         self.logger.log(logging.ERROR,getLogMessage("moveTree: %s must be a directory" % (src)))


   def finishLocalPegasusCommand(self):
      nInstanceIdDigits = 2
      instance = 1
      instanceId = str(instance).zfill(nInstanceIdDigits)

      workDirectoryName = os.path.join(self.clientData['jobPath'],'work')
      if os.path.isdir(workDirectoryName):
         shutil.rmtree(workDirectoryName,True)
      workDirectoryName = os.path.join(self.clientData['jobPath'],'scratch')
      if os.path.isdir(workDirectoryName):
         shutil.rmtree(workDirectoryName,True)
      for ftype in '.dax','_sites.xml','.pegasusrc','_tc.txt':
         pegasusFile = os.path.join(self.clientData['jobPath'],"%s_%s%s" % (self.clientData['localJobId'],instanceId,ftype))
         if os.path.isfile(pegasusFile):
            os.remove(pegasusFile)
         else:
            pegasusFile = os.path.join(self.clientData['jobPath'],"%s%s" % (self.clientData['localJobId'],ftype))
            if os.path.isfile(pegasusFile):
               os.remove(pegasusFile)

      self.moveTree(self.clientData['jobPath'],os.getcwd())
      runDirectory = os.path.join(os.getcwd(),self.clientData['runName'])
      if os.path.exists(runDirectory):
         shutil.rmtree(runDirectory,True)


   def finishLocal(self):
      signal.signal(signal.SIGCONT,self.handleContinueSignal)
      signal.signal(signal.SIGTSTP,self.handleSuspendSignal)

      if self.clientData['isParametric']:
         self.finishLocalParametric()

      for fileToRemove in self.clientData['filesToRemove']:
         if os.path.exists(fileToRemove):
            os.remove(fileToRemove)

      for fileToRemove in self.clientData['emptyFilesToRemove']:
         if os.path.exists(fileToRemove):
            if(os.path.getsize(fileToRemove) == 0):
               os.remove(fileToRemove)

      if self.clientData['isLocalPegasus']:
         self.finishLocalPegasusCommand()


   def scanLocalOutputFile(self,
                           outputFileObject):
      outChunk = os.read(outputFileObject.fileno(),1024)
      if len(outChunk) > 0:
         if self.clientData['isParametric']:
            update = False
            jobStartPattern    = re.compile('.* Executing JOB ([^ ]*) *-n *([^ ]*) *-N *([^ ]*).*/([^ ]*)')
            jobCompletePattern = re.compile('.* JOB *([0-9]+)_([0-9]+)\.sh_[A-Za-z0-9]*_[0-9]*')
            for record in outChunk.split('\n'):
               matchObj = jobStartPattern.match(record)
               if matchObj:
                  if matchObj.group(2) == matchObj.group(4):
                     instance = int(matchObj.group(3).split('_')[-1])
                     self.clientData['jobStates'][instance] = 'executing'
                     update = True
               else:
                  matchObj = jobCompletePattern.match(record)
                  if matchObj:
                     instance = int(matchObj.group(2))
                     instanceExitCode = 0
                     exitCodeFile = "%s.%s_%s" % (EXITCODE,matchObj.group(1),matchObj.group(2))
                     exitCodePath = os.path.join(self.clientData['jobPath'],matchObj.group(2),exitCodeFile)
                     if os.path.exists(exitCodePath):
                        try:
                           fpExitCode = open(exitCodePath,'r')
                           try:
                              instanceExitCode = int(fpExitCode.readline())
                           except (IOError,OSError):
                              self.logger.log(logging.ERROR,getLogMessage("%s could not be read" % (exitCodePath)))
                           finally:
                              fpExitCode.close()
                        except (IOError,OSError):
                           self.logger.log(logging.ERROR,getLogMessage("%s could not be opened" % (exitCodePath)))
                     if instanceExitCode:
                        self.clientData['jobStates'][instance] = 'failed'
                     else:
                        self.clientData['jobStates'][instance] = 'finished'
                     update = True
            if update:
               self.__updatePegasusWorkflowStatus()
         else:
            if outputFileObject == self.clientData['childReadStdout']:
               self.__writeToStdout(outChunk)
            if outputFileObject == self.clientData['childReadStderr']:
               self.__writeToStderr(outChunk)
      else:
         self.closeFile(outputFileObject)


   def checkInputsExist(self):
      inputsExist = True
      if self.clientData['operationMode'] & self.commandParser.OPERATIONMODERUNDISTRIBUTORID:
         inputFiles = []
         if self.clientData['isParametric']:
            for substitutions in self.commandParser.getNextParameterCombination():
               if self.commandParser.getOption('inputFiles'):
                  for inputFile in self.commandParser.getOption('inputFiles'):
                     template = ParameterTemplate(inputFile)
                     actualInputFile = template.substitute(substitutions).replace('@:','')
                     if not actualInputFile in inputFiles:
                        inputFiles.append(actualInputFile)
         else:
            if self.commandParser.getOption('inputFiles'):
               for inputFile in self.commandParser.getOption('inputFiles'):
                  inputFiles.append(inputFile)

         if len(inputFiles) > 0:
            for inputFile in inputFiles:
               if not os.path.exists(inputFile):
                  self.__writeToStderr("Input %s does not exist\n" % (inputFile))
                  inputsExist = False

      return(inputsExist)


   def mapInputFiles(self):
      exitCode = 0
      if self.clientData['operationMode'] & self.commandParser.OPERATIONMODERUNDISTRIBUTORID:
         inputFiles = []
         enteredCommandArguments = self.commandParser.getEnteredCommandArguments()
         if self.clientData['isParametric']:
            for substitutions in self.commandParser.getNextParameterCombination():
               if self.commandParser.getOption('inputFiles'):
                  for inputFile in self.commandParser.getOption('inputFiles'):
                     template = ParameterTemplate(inputFile)
                     actualInputFile = template.substitute(substitutions).replace('@:','')
                     if not actualInputFile in inputFiles:
                        inputFiles.append(actualInputFile)

               if len(enteredCommandArguments) > 0:
                  for arg in enteredCommandArguments:
                     template = ParameterTemplate(arg)
                     try:
                        arg = template.substitute(substitutions)
                     except KeyError,err:
                        self.__writeToStderr("Pattern substitution failed for @@%s\n" % (err[0]))
                        exitCode = 1
                     if not arg.startswith('-'):
                        actualInputFile = arg.replace('@:','')
                        if os.path.isfile(actualInputFile) or os.path.isdir(actualInputFile):
                           if not actualInputFile in inputFiles:
                              inputFiles.append(actualInputFile)
                     else:
                        lexArgs = shlex.split(arg)
                        for lexArg in lexArgs:
                           if '=' in lexArg:
                              arglets = lexArg.split('=')
                              for arglet in arglets[1:]:
                                 actualInputFile = arglet.replace('@:','')
                                 if os.path.isfile(actualInputFile) or os.path.isdir(actualInputFile):
                                    if not actualInputFile in inputFiles:
                                       inputFiles.append(actualInputFile)
                           else:
                              actualInputFile = lexArg.replace('@:','')
                              if os.path.isfile(actualInputFile) or os.path.isdir(actualInputFile):
                                 if not actualInputFile in inputFiles:
                                    inputFiles.append(actualInputFile)
         else:
            if self.commandParser.getOption('inputFiles'):
               for inputFile in self.commandParser.getOption('inputFiles'):
                  inputFiles.append(inputFile)

            if len(enteredCommandArguments) > 0:
               for arg in enteredCommandArguments:
                  if not arg.startswith('-'):
                     if os.path.isfile(arg) or os.path.isdir(arg):
                        if not arg in inputFiles:
                           inputFiles.append(arg)
                  else:
                     lexArgs = shlex.split(arg)
                     for lexArg in lexArgs:
                        if '=' in lexArg:
                           arglets = lexArg.split('=')
                           for arglet in arglets[1:]:
                              if os.path.isfile(arglet) or os.path.isdir(arglet):
                                 if not arglet in inputFiles:
                                    inputFiles.append(arglet)
                        else:
                           if os.path.isfile(lexArg) or os.path.isdir(lexArg):
                              if not lexArg in inputFiles:
                                 inputFiles.append(lexArg)

         if len(inputFiles) > 0:
            if len(self.clientData['mountPoints']) > 0:
               for inputFile in inputFiles:
                  inputPath = os.path.abspath(inputFile)
                  mountPoint = inputPath
                  while mountPoint:
                     if os.path.ismount(mountPoint):
                        try:
                           inode = os.lstat(inputFile).st_ino
                           self.clientData['mountPoints'][mountPoint]['inputFiles'].append((inputPath,inode))
                        except OSError,err:
                           self.__writeToStderr("%s: %s\n" % (inputFile,err.args[1]))
                           exitCode = err.args[0]
                        break
                     else:
                        if mountPoint != os.sep:
                           mountPoint = os.path.dirname(mountPoint)
                        else:
                           break

               markedForDeletion = []
               for mountPoint in self.clientData['mountPoints']:
                  if len(self.clientData['mountPoints'][mountPoint]['inputFiles']) == 0:
                     markedForDeletion.append(mountPoint)

               for mountPoint in markedForDeletion:
                  del self.clientData['mountPoints'][mountPoint]
               del markedForDeletion

               self.logger.log(logging.INFO,getLogMessage(self.clientData['mountPoints']))

               for mountPoint in self.clientData['mountPoints']:
                  device = self.clientData['mountPoints'][mountPoint]['device']
                  for inputFile in self.clientData['mountPoints'][mountPoint]['inputFiles']:
                     clientMessage = {'messageType':'inputFileWithMount',
                                      'path':inputFile[0],
                                      'inode':inputFile[1],
                                      'device':device}
                     self.serverConnection.postJsonMessage(clientMessage)
                     self.logger.log(logging.DEBUG,getLogMessage("inputFileWithMount = %s" % (inputFile[0])))
            else:
               for inputFile in inputFiles:
                  try:
                     inode = os.lstat(inputFile).st_ino
                     clientMessage = {'messageType':'inputFile',
                                      'path':os.path.abspath(inputFile),
                                      'inode':inode}
                     self.serverConnection.postJsonMessage(clientMessage)
                  except OSError,err:
                     self.__writeToStderr("%s: %s\n" % (inputFile,err.args[1]))
                     exitCode = err.args[0]

      if exitCode:
         self.exit(exitCode)
      else:
         clientMessage = {'messageType':'inputFileInodesSent'}
         self.serverConnection.postJsonMessage(clientMessage)


   def wait(self):
      if self.serverConnection.isConnected():
         clientMessage = {'messageType':'exit'}
         self.serverConnection.postJsonMessage(clientMessage)


   def processServerRequests(self):
      message = self.serverConnection.pullMessage(0)
      while message:
         args = message.split()
         if args[0] != 'null' and args[0] != 'json':
            self.logger.log(logging.DEBUG,getLogMessage("server request = %s" % (args[0])))

         if args[0] == 'json':
            jsonMessageLength = int(args[1])
            jsonMessage = self.serverConnection.pullMessage(jsonMessageLength)
            if len(jsonMessage) > 0:
               try:
                  serverJsonObject = json.loads(jsonMessage)
               except ValueError:
                  self.logger.log(logging.ERROR,getLogMessage("JSON object %s could not be decoded" % (jsonMessage)))
               else:
                  if serverJsonObject['messageType'] != 'null':
                     self.logger.log(logging.DEBUG,getLogMessage("server request = %s" % (serverJsonObject['messageType'])))
                  if   serverJsonObject['messageType'] == 'serverVersion':
                     self.serverData['version'] = serverJsonObject['version']
                  elif serverJsonObject['messageType'] == 'message':
                     self.logger.log(logging.INFO,getLogMessage(serverJsonObject['text']))
                  elif serverJsonObject['messageType'] == 'authz':
                     self.serverData['authz'] = serverJsonObject['success']
                     if not self.serverData['authz']:
                        self.logger.log(logging.ERROR,getLogMessage("Session authentication failed"))
                        self.clientIdAuthAttributes.clearSessionToken()
                        tryAgain = serverJsonObject['retry']
                        if tryAgain:
                           self.getUserAttributes()
                           if self.haveIdAuthAttributes():
                              self.signon()
                     else:
                        if   self.clientData['authorizedAction'] == self.AUTHORIZEDACTIONEXECUTE:
                           if self.clientData['attachId'] == 0L:
                              self.mapSubmitCommandFiles()
                           else:
                              clientMessage = {'messageType':'parseArguments'}
                              self.serverConnection.postJsonMessage(clientMessage)
                        elif self.clientData['authorizedAction'] == self.AUTHORIZEDACTIONHEARTBEAT:
                           clientMessage = {'messageType':'clientHeartbeat'}
                           self.serverConnection.postJsonMessage(clientMessage)
                           self.scheduleHeartbeat()
                        elif self.clientData['authorizedAction'] == self.AUTHORIZEDACTIONSENDMETRICS:
                           self.finishLocal()
                           clientMessage = {'messageType':'localExit',
                                            'status':self.clientData['exitCode'],
                                            'cpuTime':self.clientData['cpuTime'],
                                            'realTime':self.clientData['realTime']}
                           self.serverConnection.postJsonMessage(clientMessage)
                           self.clientData['readyToExit'] = True

                        self.clientData['authorizedAction'] == self.AUTHORIZEDACTIONNONE
                  elif serverJsonObject['messageType'] == 'writeStdout':
                     self.__writeToStdout(serverJsonObject['text'])
                  elif serverJsonObject['messageType'] == 'writeStderr':
                     self.__writeToStderr(serverJsonObject['text'])
                  elif serverJsonObject['messageType'] == 'write':
                     serverOutputFile = serverJsonObject['file']
                     text = serverJsonObject['text']
                     if len(text) > 0:
                        found = False
                        if not found:
                           for importObject in self.clientData['importObjects']:
                              if self.clientData['importObjects'][importObject]['path'] == serverOutputFile:
                                 self.clientData['importObjects'][importObject]['buffer'] += text
                                 found = True
                                 break
                        if not found:
                           self.logger.log(logging.ERROR,getLogMessage("File object not found to match %s" % (serverOutputFile)))
                  elif serverJsonObject['messageType'] == 'close':
                     serverOutputFile = serverJsonObject['file']
                     found = False
                     for importObject in self.clientData['importObjects']:
                        if self.clientData['importObjects'][importObject]['path'] == serverOutputFile:
                           self.clientData['closePendingObjects'].append(importObject)
                           found = True
                           break
                     if not found:
                        self.logger.log(logging.ERROR,getLogMessage("File object not found to match %s" % (serverOutputFile)))
                  elif serverJsonObject['messageType'] == 'jobId':
                     self.serverData['jobId'] = serverJsonObject['jobId']
                     if self.serverData['jobId'] == 0:
                        self.exit(1)
                     else:
                        self.clientData['localJobId'] = "%08d" % (self.serverData['jobId'])
                        if not self.clientData['runName']:
                           self.clientData['runName'] = self.clientData['localJobId']
                     if self.clientData['reportMetrics']:
                        self.__writeToStderr("=SUBMIT-METRICS=> job=%s\n" % (self.clientData['localJobId']))
                  elif serverJsonObject['messageType'] == 'jobToken':
                     self.serverData['jobToken'] = serverJsonObject['token']
                  elif serverJsonObject['messageType'] == 'submitterClass':
                     self.serverData['submitterClass'] = serverJsonObject['class']
                  elif serverJsonObject['messageType'] == 'runName':
                     self.clientData['runName']    = serverJsonObject['runName']
                     self.serverData['jobId']      = serverJsonObject['jobId']
                     self.clientData['localJobId'] = "%08d" % (self.serverData['jobId'])
                  elif serverJsonObject['messageType'] == 'event':
                     self.serverData['event'] = serverJsonObject['text']
                  elif serverJsonObject['messageType'] == 'operationMode':
                     self.clientData['operationMode'] = serverJsonObject['operationMode']
                  elif serverJsonObject['messageType'] == 'createdSessions':
                     self.serverData['createdSessions'] = serverJsonObject['createdSessions']
                  elif serverJsonObject['messageType'] == 'heartbeatInterval':
                     self.serverData['heartbeatInterval'] = serverJsonObject['interval']
                  elif serverJsonObject['messageType'] == 'pegasusRC':
                     self.serverData['pegasusRC'] = serverJsonObject['text']
                  elif serverJsonObject['messageType'] == 'pegasusSites':
                     self.serverData['pegasusSites'] = serverJsonObject['text']
                  elif serverJsonObject['messageType'] == 'submissionScripts':
                     submissionScripts = serverJsonObject['submissionScripts']
                     self.serverData['submissionScripts'] = SubmissionScriptsInfo('Client',
                                                                                  submissionScripts=submissionScripts)
                  elif serverJsonObject['messageType'] == 'clientSuspended':
                     self.clientData['suspended'] = True
                     signal.signal(signal.SIGCONT,self.handleContinueSignal)
                     signal.signal(signal.SIGTSTP,signal.SIG_DFL)
                     os.kill(os.getpid(),signal.SIGTSTP)
                  elif serverJsonObject['messageType'] == 'clientContinued':
                     self.clientData['suspended'] = False
                     signal.signal(signal.SIGCONT,signal.SIG_DFL)
                     signal.signal(signal.SIGTSTP,self.handleSuspendSignal)
                     os.kill(os.getpid(),signal.SIGCONT)
                  elif serverJsonObject['messageType'] == 'jobScanUpdate':
                     messageText = serverJsonObject['text']
                     if self.clientData['jobScanPath'] == "":
                        jobScanPath = os.path.join(self.clientData['workDirectory'],self.clientData['runName'])
                        if not os.path.exists(jobScanPath):
                           os.makedirs(jobScanPath)
                        jobScanPath = os.path.join(jobScanPath,'parameterCombinations.csv')
                        self.clientData['jobScanPath'] = jobScanPath
                     try:
                        parameterCombinationsDir  = os.path.dirname(self.clientData['jobScanPath'])
                        parameterCombinationsBase = os.path.basename(self.clientData['jobScanPath'])
                        if '.' in parameterCombinationsBase:
                           parameterCombinationsBase = parameterCombinationsBase.split('.')[0]
                        tmpParameterCombinationsFile = parameterCombinationsBase + '.tmp'
                        tmpParameterCombinationsPath = os.path.join(parameterCombinationsDir,tmpParameterCombinationsFile)
                        renameTmpFile = False
                        try:
                           fpJobScan = open(tmpParameterCombinationsPath,'w')
                           try:
                              fpJobScan.write(messageText)
                              renameTmpFile = True
                           except (IOError,OSError):
                              self.logger.log(logging.ERROR,getLogMessage("%s could not be written" % \
                                                                          (tmpParameterCombinationsPath)))
                           finally:
                              fpJobScan.close()
                        except (IOError,OSError):
                           self.logger.log(logging.ERROR,getLogMessage("%s could not be opened" % (tmpParameterCombinationsPath)))
                        if renameTmpFile:
                           os.rename(tmpParameterCombinationsPath,self.clientData['jobScanPath'])
                     except:
                        pass
                     if self.clientData['progressReport'] == 'curses':
                        if os.path.exists(self.clientData['jobScanPath']):
                           if not self.jobScanner and not self.clientData['jobScannerStarted']:
                              self.jobScanner = JobScanner(self.clientData['jobScanPath'])
                              self.clientData['jobScannerStarted'] = True
                              if not self.jobScanner.start():
                                 self.jobScanner.finish()
                                 self.jobScanner = None
                  elif serverJsonObject['messageType'] == 'runStatusUpdate':
                     messageText = serverJsonObject['text']
                     saveFile    = serverJsonObject['saveFile']
                     if self.clientData['runStatusPath'] == "":
                        runStatusPath = os.path.join(self.clientData['workDirectory'],self.clientData['runName'])
                        if saveFile and not os.path.exists(runStatusPath):
                           os.makedirs(runStatusPath)
                        runStatusPath = os.path.join(runStatusPath,'pegasusstatus.txt')
                        self.clientData['runStatusPath'] = runStatusPath
                     try:
                        if saveFile:
                           runStatusDir  = os.path.dirname(self.clientData['runStatusPath'])
                           runStatusBase = os.path.basename(self.clientData['runStatusPath'])
                           if '.' in runStatusBase:
                              runStatusBase = runStatusBase.split('.')[0]
                           tmpRunStatusFile = runStatusBase + '.tmp'
                           tmpRunStatusPath = os.path.join(runStatusDir,tmpRunStatusFile)
                           renameTmpFile = False
                           try:
                              fpJobScan = open(tmpRunStatusPath,'w')
                              try:
                                 fpJobScan.write(messageText)
                                 renameTmpFile = True
                              except (IOError,OSError):
                                 self.logger.log(logging.ERROR,getLogMessage("%s could not be written" % (tmpRunStatusPath)))
                              finally:
                                 fpJobScan.close()
                           except (IOError,OSError):
                              self.logger.log(logging.ERROR,getLogMessage("%s could not be opened" % (tmpRunStatusPath)))
                           if renameTmpFile:
                              os.rename(tmpRunStatusPath,self.clientData['runStatusPath'])
                     except:
                        pass
                  elif serverJsonObject['messageType'] == 'serverReadyForIO':
                     if self.clientData['localExecution']:
                        self.startLocal()
                     else:
                        self.clientData['inputObjects'][sys.stdin] = '#STDIN#'
                        if self.clientData['detach']:
                           clientMessage = {'messageType':'detachSignal'}
                           self.serverConnection.postJsonMessage(clientMessage)
                  elif serverJsonObject['messageType'] == 'readyToDetach':
                     clientMessage = {'messageType':'detach'}
                     self.serverConnection.postJsonMessage(clientMessage)
                     if self.jobScanner:
                        self.jobScanner.finish()
                        self.jobScanner = None
                     self.clientData['readyToExit'] = True
                     self.clientData['exitCode']    = 0
                     notificationMessage = \
                                     "Detaching from run %s.\n" % (self.clientData['runName']) + \
                                     "    Check run status with the command: submit --status %d\n" % (self.serverData['jobId']) + \
                                     "       Terminate run with the command: submit --kill %d\n" % (self.serverData['jobId']) + \
                                     "   Re-connect to run with the command: submit --attach %d\n" % (self.serverData['jobId'])
                     self.__writeToStdout(notificationMessage)
                  elif serverJsonObject['messageType'] == 'attached':
                     notificationMessage = "Attached to run %d.\n" % (self.clientData['attachId'])
                     self.__writeToStdout(notificationMessage)
                     clientMessage = {'messageType':'clientVersion',
                                      'version':self.clientData['version']}
                     self.serverConnection.postJsonMessage(clientMessage)
                  elif serverJsonObject['messageType'] == 'serverReadyForInputMapping':
                     self.mapInputFiles()
                  elif serverJsonObject['messageType'] == 'addExportFile':
                     self.clientData['transferFiles'].append(serverJsonObject['file'])
                  elif serverJsonObject['messageType'] == 'exportFiles':
                     transferTarFile = serverJsonObject['file']
                     transferTarPath = os.path.join(os.sep,"tmp",transferTarFile)
                     tarCommand = buildCreateTarCommand(transferTarPath,self.clientData['transferFiles'],useAbsolutePaths=True)
                     self.logger.log(logging.INFO,getLogMessage("command = " + str(tarCommand)))
                     try:
                        child = subprocess.Popen(tarCommand,
                                                 stdout=subprocess.PIPE,
                                                 stderr=subprocess.PIPE,
                                                 close_fds=True)
                        tarStdOutput,tarStdError = child.communicate()
                        tarExitStatus = child.returncode
                        if tarExitStatus != 0:
                           self.logger.log(logging.ERROR,getLogMessage("Failed to write transfer tarfile %s" % (transferTarPath)))
                           if tarStdOutput:
                              self.logger.log(logging.ERROR,getLogMessage(tarStdOutput))
                           if tarStdError:
                              self.logger.log(logging.ERROR,getLogMessage(tarStdError))
                           exitCode = 1
                        else:
                           try:
                              fp = open(transferTarPath,'r')
                           except (IOError,OSError):
                              self.logger.log(logging.ERROR,getLogMessage("%s could not be opened" % (transferTarPath)))
                           else:
                              self.clientData['exportObjects'][fp] = transferTarFile
                     except OSError,err:
                        self.logger.log(logging.ERROR,getLogMessage("Failed to to create transfer tarfile %s" % (transferTarPath)))
                        self.logger.log(logging.ERROR,getLogMessage(err.args[1]))
                  elif serverJsonObject['messageType'] == 'exportFilesComplete':
                     clientMessage = {'messageType':'startRemote'}
                     self.serverConnection.postJsonMessage(clientMessage)
                  elif serverJsonObject['messageType'] == 'addExportCommandFile':
                     self.clientData['commandFiles'].append(serverJsonObject['file'])
                  elif serverJsonObject['messageType'] == 'exportCommandFiles':
                     transferTarFile = serverJsonObject['file']
                     transferTarPath = os.path.join(os.sep,"tmp",transferTarFile)
                     tarCommand = buildCreateTarCommand(transferTarPath,self.clientData['commandFiles'],useAbsolutePaths=True)
                     self.logger.log(logging.INFO,getLogMessage("command = " + str(tarCommand)))
                     try:
                        child = subprocess.Popen(tarCommand,
                                                 stdout=subprocess.PIPE,
                                                 stderr=subprocess.PIPE,
                                                 close_fds=True)
                        tarStdOutput,tarStdError = child.communicate()
                        tarExitStatus = child.returncode
                        if tarExitStatus != 0:
                           self.logger.log(logging.ERROR,getLogMessage("Failed to write transfer tarfile %s" % (transferTarPath)))
                           if tarStdOutput:
                              self.logger.log(logging.ERROR,getLogMessage(tarStdOutput))
                           if tarStdError:
                              self.logger.log(logging.ERROR,getLogMessage(tarStdError))
                           exitCode = 1
                        else:
                           try:
                              fp = open(transferTarPath,'r')
                           except (IOError,OSError):
                              self.logger.log(logging.ERROR,getLogMessage("%s could not be opened" % (transferTarPath)))
                           else:
                              self.clientData['exportObjects'][fp] = transferTarFile
                     except OSError,err:
                        self.logger.log(logging.ERROR,getLogMessage("Failed to to create transfer tarfile %s" % (transferTarPath)))
                        self.logger.log(logging.ERROR,getLogMessage(err.args[1]))
                  elif serverJsonObject['messageType'] == 'exportCommandFilesComplete':
                     clientMessage = {'messageType':'parseArguments'}
                     self.serverConnection.postJsonMessage(clientMessage)
                  elif serverJsonObject['messageType'] == 'argumentsParsed':
                     if   self.clientData['localExecution']:
                        clientMessage = {'messageType':'startLocal'}
                        self.serverConnection.postJsonMessage(clientMessage)
                     elif not (self.clientData['operationMode'] & self.commandParser.OPERATIONMODERUNPROXY):
                        clientMessage = {'messageType':'setupRemote'}
                        self.serverConnection.postJsonMessage(clientMessage)
                  elif serverJsonObject['messageType'] == 'childHasExited':
                     self.serverData['childHasExited'] = serverJsonObject['childHasExited']
                     if not self.clientData['importFilesComplete']:
                        clientMessage = {'messageType':'clientReadyForIO'}
                        self.serverConnection.postJsonMessage(clientMessage)
                  elif serverJsonObject['messageType'] == 'noImportFile':
                     self.clientData['importFilesComplete'] = True
                     clientMessage = {'messageType':'importFilesComplete'}
                     self.serverConnection.postJsonMessage(clientMessage)
                  elif serverJsonObject['messageType'] == 'importFile':
                     transferTarFile = serverJsonObject['file']
                     transferTarPath = os.path.join(self.clientData['workDirectory'],transferTarFile)
                     try:
                        fp = open(transferTarPath,'w')
                     except (IOError,OSError):
                        self.logger.log(logging.ERROR,getLogMessage("Failed to to create transfer tarfile %s" % \
                                                                                             (transferTarPath)))
                     else:
                        self.clientData['importObjects'][fp]           = {}
                        self.clientData['importObjects'][fp]['path']   = transferTarFile
                        self.clientData['importObjects'][fp]['buffer'] = ""
                        clientMessage = {'messageType':'importFileReady',
                                         'file':transferTarFile}
                        self.serverConnection.postJsonMessage(clientMessage)
                  elif serverJsonObject['messageType'] == 'importFileFailed':
                     transferTarFile = serverJsonObject['file']
                     fpDel = None
                     for fp in self.clientData['importObjects']:
                        if self.clientData['importObjects'][fp]['path'] == transferTarFile:
                           fpDel = fp
                           break
                     if fpDel:
                        del self.clientData['importObjects'][fpDel]
                     self.clientData['importFilesComplete'] = True
                     clientMessage = {'messageType':'importFilesComplete'}
                     self.serverConnection.postJsonMessage(clientMessage)
                  elif serverJsonObject['messageType'] == 'serverExit':
                     self.serverData['exited'] = True
                     exitCode = serverJsonObject['exitCode']
                     self.clientData['readyToExit'] = True
                     self.clientData['exitCode']    = exitCode
                     self.exit(exitCode)
                  elif serverJsonObject['messageType'] == 'exit':
                     exitCode = serverJsonObject['exitCode']
                     self.exit(exitCode)
                  elif serverJsonObject['messageType'] == 'wait':
                     self.wait()
                  elif serverJsonObject['messageType'] == 'null':
                     pass
                  else:
                     self.logger.log(logging.ERROR,getLogMessage("Discarded message type: %s" % \
                                                                  (serverJsonObject['messageType'])))
            else:
               self.serverConnection.pushMessage(message + '\n')
               break
         else:
            self.logger.log(logging.ERROR,getLogMessage("Discarded server message: %s" % (message)))

         message = self.serverConnection.pullMessage(0)


   def run(self):
      while True:
         try:
            serverReader,inputObjects,exportObjects,childProcessInputObjects = self.getInputObjects()
            serverWriter,importObjects                                       = self.getOutputObjects()
            if self.clientData['childHasExited'] or self.clientData['readyToExit']:
               selectTimeout = 1.
            elif self.jobScanner:
               selectTimeout = 10.
            else:
               minSelectTimeout = self.clientData['selectTimeout']* \
                                  (1.-self.clientData['selectTimeoutSpread']/2.)
               maxSelectTimeout = self.clientData['selectTimeout']* \
                                  (1.+self.clientData['selectTimeoutSpread']/2.)
               selectTimeout = random.uniform(minSelectTimeout,maxSelectTimeout)

            readyReaders,readyWriters,readyExceptions = select.select(serverReader+inputObjects+
                                                                                   exportObjects+childProcessInputObjects,
                                                                      serverWriter+importObjects,[],
                                                                      selectTimeout)
#           if not readyReaders and not readyWriters:
#              if not self.clientData['readyToExit']:
#                 clientMessage = {'messageType':'null'}
#                 self.serverConnection.postJsonMessage(clientMessage)
         except select.error,err:
            readyReaders = []
            readyWriters = []
#
# send keepalive to notify server client is still running
#
         if self.serverConnection and self.serverConnection.isConnected():
            if not self.clientData['readyToExit']:
               timeSinceLastWrite = time.time() - self.serverConnection.getConnectionWriteTime()
               if timeSinceLastWrite >= self.clientData['selectTimeout']:
                  clientMessage = {'messageType':'null'}
                  self.serverConnection.postJsonMessage(clientMessage)
#
# If the timeout has occurred (nothing to read/write)
#
         if readyReaders == [] and readyWriters == []:
            if   self.jobScanner:
               wantToQuit,jobsFinished = self.jobScanner.processInput()
               if wantToQuit:
                  self.jobScanner.finish()
                  self.jobScanner = None
                  if self.clientData['localExecution']:
                     resultsPath = self.clientData['jobPath']
                  else:
                     resultsPath = os.path.join(self.clientData['jobPath'],self.clientData['runName'])
                  self.__writeToStdout("Simulations complete. Results are stored in directory %s\n" % (resultsPath))
            elif self.clientData['readyToExit']:
               self.exit()

         for readyReader in readyReaders:
            if   readyReader in serverReader:
               self.serverConnection.receiveMessage()
            elif readyReader in childProcessInputObjects:
               self.scanLocalOutputFile(readyReader)
            elif self.jobScanner and readyReader == sys.stdin:
               wantToQuit,jobsFinished = self.jobScanner.processInput()
               if wantToQuit:
                  self.jobScanner.finish()
                  self.jobScanner = None
                  if self.clientData['localExecution']:
                     resultsPath = self.clientData['jobPath']
                  else:
                     resultsPath = os.path.join(self.clientData['jobPath'],self.clientData['runName'])
                  self.__writeToStdout("Simulations complete. Results are stored in directory %s\n" % (resultsPath))
                  if not jobsFinished:
                     os.kill(os.getpid(),signal.SIGINT)
            else:
               self.readFile(readyReader)

         self.processServerRequests()

         for readyWriter in readyWriters:
            if   readyWriter in serverWriter:
               self.serverConnection.sendMessage()
            elif readyWriter in importObjects:
               self.writeFile(readyWriter)
         self.checkClosePendingObjects()

         serverDisconnected = False
         if   self.serverConnection and self.serverConnection.isConnected():
            lastIOTimes = []
            lastIOTimes.append(self.serverConnection.getConnectionReadTime())
#           lastIOTimes.append(self.serverConnection.getConnectionWriteTime())
            if lastIOTimes:
               timeSinceLastIO = time.time() - max(lastIOTimes)
               if self.serverConnection and self.serverConnection.isConnected():
                  if timeSinceLastIO > self.clientData['disconnectMax']:
                     if not self.clientData['suspended']:
                        serverDisconnected = True
         elif self.serverConnection:
            if not self.clientData['localExecution']:
               if not self.serverData['exited']:
                  serverDisconnected = True

         if serverDisconnected:
            self.logger.log(logging.DEBUG,getLogMessage("server disconnected"))
            # equivalent to 'readyToDetach'
            if self.jobScanner:
               self.jobScanner.finish()
               self.jobScanner = None
            self.clientData['readyToExit'] = True
            self.clientData['exitCode']    = 0
            notificationMessage = \
                            "Detaching from run %s.\n" % (self.clientData['runName']) + \
                            "    Check run status with the command: submit --status %d\n" % (self.serverData['jobId']) + \
                            "       Terminate run with the command: submit --kill %d\n" % (self.serverData['jobId']) + \
                            "   Re-connect to run with the command: submit --attach %d\n" % (self.serverData['jobId'])
            self.__writeToStdout(notificationMessage)

         if self.clientData['childPid']:
            pid = 0
            try:
               (pid,exitCode) = os.waitpid(self.clientData['childPid'],os.WNOHANG)
               if exitCode != 0:
                  if   os.WIFSIGNALED(exitCode):
                     exitCode = (1 << 7) | os.WTERMSIG(exitCode)
                  else:
                     if os.WIFEXITED(exitCode):
                        exitCode = os.WEXITSTATUS(exitCode)
            except:
               try:
                  os.kill(self.clientData['childPid'],0)
               except:
                  pid = self.clientData['childPid']
                  exitCode = 1

            if pid != 0:
               self.clientData['childPid']       = None
               self.clientData['exitCode']       = exitCode
               self.clientData['childHasExited'] = True

               rusage = resource.getrusage(resource.RUSAGE_CHILDREN)
               self.clientData['localFinishTime'] = time.time()
               self.clientData['realTime'] = self.clientData['localFinishTime']-self.clientData['localStartTime']
               self.clientData['cpuTime']  = rusage[0] + rusage[1]
               self.clientData['finishDate'] = time.strftime("%a %b %e %X %Z %Y")
               if self.clientData['isParametric']:
                  self.__updatePegasusWorkflowStatus()

               if self.clientData['reportMetrics']:
                  message =  "=SUBMIT-METRICS=>"
                  message += " job=%s" % (self.clientData['localJobId'])
                  message += " venue=local"
                  message += " status=%d" % (self.clientData['exitCode'])
                  message += " cputime=%f" % (self.clientData['cpuTime'])
                  message += " realtime=%f" % (self.clientData['realTime'])
                  self.__writeToStderr("%s\n" % (message))

               self.connect()
               if self.clientData['workDirectoryDevice']:
                  clientMessage = {'messageType':'pwdWithMount',
                                   'path':self.clientData['workDirectory'],
                                   'inode':self.clientData['workDirectoryInode'],
                                   'device':self.clientData['workDirectoryDevice']}
                  self.serverConnection.postJsonMessage(clientMessage)
               else:
                  clientMessage = {'messageType':'pwd',
                                   'path':self.clientData['workDirectory'],
                                   'inode':self.clientData['workDirectoryInode']}
                  self.serverConnection.postJsonMessage(clientMessage)
               clientMessage = {'messageType':'jobId',
                                'jobId':self.serverData['jobId']}
               self.serverConnection.postJsonMessage(clientMessage)
               clientMessage = {'messageType':'jobToken',
                                'token':self.serverData['jobToken']}
               self.serverConnection.postJsonMessage(clientMessage)
               clientMessage = {'messageType':'operationMode',
                                'operationMode':self.clientData['operationMode']}
               self.serverConnection.postJsonMessage(clientMessage)
               clientMessage = {'messageType':'createdSessions',
                                'createdSessions':self.serverData['createdSessions']}
               self.serverConnection.postJsonMessage(clientMessage)

               try:
                  submitApplicationRevision = os.environ["SUBMIT_APPLICATION_REVISION"]
               except:
                  submitApplicationRevision = ""
               if submitApplicationRevision != "":
                  clientMessage = {'messageType':'event',
                                   'text':submitApplicationRevision}
                  self.serverConnection.postJsonMessage(clientMessage)
               else:
                  clientMessage = {'messageType':'event',
                                   'text':self.serverData['event']}
                  self.serverConnection.postJsonMessage(clientMessage)

               self.cancelHeartbeat()
               self.clientData['authorizedAction'] = self.AUTHORIZEDACTIONSENDMETRICS
               self.signon()


